#!/bin/bash 

# This makes sure we exit if anything unexpected happens.
set -e

# Check for the SDK.
if [ -z "$OUTRUN_SDK_PATH" ]
then
  echo OUTRUN_SDK_PATH environment variable not set!
  exit 1
fi

# Set GCC version, if not defined externally.
if [ -z "$OUTRUN_GCC_VERSION" ]
then
  OUTRUN_GCC_VERSION=8.2.0
fi

if [ -z "$OUTRUN_GCC_PATH" ]
then
  if [ -e "$OUTRUN_SDK_PATH/../gcc-$OUTRUN_GCC_VERSION" ]
  then
    # try gcc-<version>
    OUTRUN_GCC_PATH=$(cd $OUTRUN_SDK_PATH/../gcc-$OUTRUN_GCC_VERSION; pwd)
    echo Using $OUTRUN_GCC_PATH
  else
    echo OUTRUN_GCC_PATH environment variable not set!
    exit 1 
  fi
fi

OUTRUN_SDK_INCLUDE=$OUTRUN_SDK_PATH/include
OUTRUN_SDK_LDSCRIPT=$OUTRUN_SDK_PATH/ldscript
OUTRUN_SDK_LIB=$OUTRUN_SDK_PATH/lib
OUTPUT_PATH=output

OUTRUN_GCC_ARCH=m68k-elf
OUTRUN_GCC_PREFIX=$OUTRUN_GCC_ARCH-
OUTRUN_GCC_CPUTYPE=m68000
OUTRUN_GCC_PREFIX_PATH="$OUTRUN_GCC_PATH/bin/$OUTRUN_GCC_PREFIX"

# CPU type subpath. If we have multi-libs (pre GCC 8), we need to specify /m68000 to pick the correct startup/libraries.
CPU_PATH=
if [ -d "$OUTRUN_GCC_PATH/lib/gcc/$OUTRUN_GCC_ARCH/$OUTRUN_GCC_VERSION/$OUTRUN_GCC_CPUTYPE" ]; then
  CPU_PATH=/$OUTRUN_GCC_CPUTYPE
fi

# Fixme! This may clash with a compiler already in our path.
export PATH=$PATH:$OUTRUN_GCC_PATH/bin
export PATH=$PATH:$OUTRUN_GCC_PATH/../bin

# Collect input files.
declare -a inputfiles0 inputfiles1 sharedfiles
for file in *.c *.cpp *.s ../common/*.c;
do
  if [ -e "$file" ]; then 
    # Check in what list(s) to put this file.
    if [[ $file == *"sub"* ]]; then 
	  inputfiles1+=($file);
	elif [[ $file == *"shared"* ]]; then 
	  sharedfiles+=($file);
	else
	  inputfiles0+=($file);
	fi
  fi # file exists
done

# Make output path.
if [ "$1" != "clean" ]; then
  if ! [ -e "$OUTPUT_PATH" ]; then mkdir $OUTPUT_PATH; fi

  # compile our files.
  echo Compiling using GCC version $OUTRUN_GCC_VERSION...
fi

# Clean out linker scripts.
if [ -e "$OUTPUT_PATH/main.link.in" ]; then rm "$OUTPUT_PATH/main.link.in"; fi
if [ -e "$OUTPUT_PATH/sub.link.in" ]; then rm "$OUTPUT_PATH/sub.link.in"; fi

# Build CPU1 only if there are specific files for it.
buildcpus="0 1"
if [ ${#inputfiles1[@]} -eq 0 ]; then
  buildcpus="0"
fi

for cpu in $buildcpus; do
  filelists="inputfiles$cpu sharedfiles"
  for filelist in $filelists; do
    eval contents=\( \${${filelist}[@]} \)
	for file in ${contents[@]}; do
	  
	  # For every file, figure out the output name.
      filename=$(basename "$file")
      basename="${filename%.*}"
	  if [ $filelist == "sharedfiles" ]; then
	    objname="$basename"_cpu"$cpu".o
	  else
	    objname="$basename".o
	  fi
      objfile="$OUTPUT_PATH"/"$objname"

      if [ "$1" = "clean" ]; then
	    # Delete output file.
	    if [ -e "$objfile" ]; then rm "$objfile"; fi
	  else
	    # Compile/assemble file.
	    echo $filename
	    extension="${filename##*.}"

        if [ $extension == "c" ]; then
          "$OUTRUN_GCC_PREFIX_PATH""gcc" -c $file -std=gnu11 -$OUTRUN_GCC_CPUTYPE -o "$objfile" -Os "-D""CPU$cpu" -I. -I$OUTRUN_SDK_INCLUDE -I"$OUTRUN_SDK_INCLUDE"/cpu$cpu -I../common
        fi # c

        if [ $extension == "cpp" ]; then
          "$OUTRUN_GCC_PREFIX_PATH""g++" -c $file --no-rtti -$OUTRUN_GCC_CPUTYPE -o "$objfile" -Os "-D""CPU$cpu" -I. -I$OUTRUN_SDK_INCLUDE -I"$OUTRUN_SDK_INCLUDE"/cpu$cpu -I../common
        fi # cpp

        if [ $extension == "s" ]; then
          "$OUTRUN_GCC_PREFIX_PATH""as" $file -$OUTRUN_GCC_CPUTYPE -o "$objfile" --defsym "CPU$cpu=1"
        fi # s
		
        # Append to linker input list
        if [ $cpu == 1 ]; then echo $objfile >> "$OUTPUT_PATH/sub.link.in"; fi
        if [ $cpu == 0 ]; then echo $objfile >> "$OUTPUT_PATH/main.link.in"; fi
		
      fi # compile
	  
	done # file
  done # file list
done # cpu

if [ "$1" = "clean" ] 
then
  for file in epr-10380b.133 epr-10382b.118 epr-10381b.132 epr-10383b.117 epr-10327a.76 epr-10329a.58 epr-10328a.75 epr-10330a.57 main.link.in sub.link.in maincpu_ram.bin maincpu_ram.map maincpu_rom.bin maincpu_rom.map subcpu_ram.bin subcpu_ram.map subcpu_rom.bin subcpu_rom.map
  do
    if [ -e "$OUTPUT_PATH/$file" ]; then rm "$OUTPUT_PATH/$file"; fi
  done
  exit 0 # Done cleaning.
fi

# link
echo Linking...
echo maincpu_rom.bin
# crti.o crtbegin.o ... -lgcc crtend.o crtn.o

"$OUTRUN_GCC_PATH/bin/""$OUTRUN_GCC_PREFIX""ld" "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtbegin.o" @$OUTPUT_PATH/main.link.in $OUTRUN_SDK_LIB/outrun_sdk_cpu0.lib -lc -lgcc "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtend.o" -L"$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH" -L"$OUTRUN_GCC_PATH/m68k-elf/lib$CPU_PATH" --script=$OUTRUN_SDK_LDSCRIPT/outrun_main_rom.ld -o $OUTPUT_PATH/maincpu_rom.bin --Map=$OUTPUT_PATH/maincpu_rom.map
# if ERRORLEVEL 1 goto error
echo maincpu_ram.bin
"$OUTRUN_GCC_PATH/bin/""$OUTRUN_GCC_PREFIX""ld" "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtbegin.o" @$OUTPUT_PATH/main.link.in $OUTRUN_SDK_LIB/outrun_sdk_cpu0.lib -lc -lgcc "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtend.o" -L"$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH" -L"$OUTRUN_GCC_PATH/m68k-elf/lib$CPU_PATH" --script=$OUTRUN_SDK_LDSCRIPT/outrun_main_ram.ld -o $OUTPUT_PATH/maincpu_ram.bin --Map=$OUTPUT_PATH/maincpu_ram.map
# if ERRORLEVEL 1 goto error

if [ -e "$OUTPUT_PATH""/sub.link.in" ]
then
  echo subcpu_rom.bin
  "$OUTRUN_GCC_PATH/bin/""$OUTRUN_GCC_PREFIX""ld" "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtbegin.o" @$OUTPUT_PATH/sub.link.in $OUTRUN_SDK_LIB/outrun_sdk_cpu1.lib -lc -lgcc "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtend.o" -L"$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH" -L"$OUTRUN_GCC_PATH/m68k-elf/lib$CPU_PATH" --script=$OUTRUN_SDK_LDSCRIPT/outrun_sub_rom.ld -o $OUTPUT_PATH/subcpu_rom.bin --Map=$OUTPUT_PATH/subcpu_rom.map
# if ERRORLEVEL 1 goto error
  echo subcpu_ram.bin
  "$OUTRUN_GCC_PATH/bin/""$OUTRUN_GCC_PREFIX""ld" "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtbegin.o" @$OUTPUT_PATH/sub.link.in $OUTRUN_SDK_LIB/outrun_sdk_cpu1.lib -lc -lgcc "$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH/crtend.o" -L"$OUTRUN_GCC_PATH/lib/gcc/m68k-elf/$OUTRUN_GCC_VERSION$CPU_PATH" -L"$OUTRUN_GCC_PATH/m68k-elf/lib$CPU_PATH" --script=$OUTRUN_SDK_LDSCRIPT/outrun_sub_ram.ld -o $OUTPUT_PATH/subcpu_ram.bin --Map=$OUTPUT_PATH/subcpu_ram.map
# if ERRORLEVEL 1 goto error
fi

# delete linker input lists
if [ -e "$OUTPUT_PATH\main.link.in" ]; then rm "$OUTPUT_PATH\main.link.in"; fi
if [ -e "$OUTPUT_PATH\sub.link.in" ]; then rm "$OUTPUT_PATH\sub.link.in"; fi

# build rom images.
echo Building ROM images...

"$OUTRUN_SDK_PATH/../bin/"splitbin "$OUTPUT_PATH/maincpu_rom.bin" 65536 2 "$OUTPUT_PATH/epr-10380b.133" "$OUTPUT_PATH/epr-10382b.118" "$OUTPUT_PATH!/epr-10381b.132" "$OUTPUT_PATH/epr-10383b.117"
# if ERRORLEVEL 1 goto error

if [ -e "$OUTPUT_PATH/subcpu_rom.bin" ] 
then 
  "$OUTRUN_SDK_PATH/../bin/"splitbin "$OUTPUT_PATH/subcpu_rom.bin" 65536 2 "$OUTPUT_PATH/epr-10327a.76" "$OUTPUT_PATH/epr-10329a.58" "$OUTPUT_PATH/epr-10328a.75" "$OUTPUT_PATH/epr-10330a.57"
#  if ERRORLEVEL 1 goto error
fi

echo Done.
