#include "hwinit.h"
#include "text.h"
#include "tile.h"
#include "palette.h"
#include "input.h"
#include "menu.h"
#include <irq.h>
#include <stdint.h>
#include <io.h>
#include "orsound.h"
#include "tileunpack.h"
#include <sprite.h>
#include "paletteutil.h"
#include <road.h>
#include <stdlib.h>
#include <memory.h>

// These don't really fit on the screen anymore.
//#define HAVE_MUTE_OPTIONS 1

// Display play time (hex) on top of the screen. Used to build the TrackInfo list.
//#define DEBUG_PLAY_TIME 1

extern const TileGraphics RadioGraphics;

void SetupGradient ()
{
	const uint8_t colorStart[3] = { 0x00, 0x40, 0x80 };
	const uint8_t colorEnd[3]   = { 0x40, 0xe0, 0xf0 };
	
	uint16_t red   = (uint16_t)colorStart[0] << 8;
	uint16_t green = (uint16_t)colorStart[1] << 8;
	uint16_t blue  = (uint16_t)colorStart[2] << 8;
	
	int16_t d_red   = (colorEnd[0] - colorStart[0]) << 1;
	int16_t d_green = (colorEnd[1] - colorStart[1]) << 1;
	int16_t d_blue  = (colorEnd[2] - colorStart[2]) << 1;
	
	int16_t err_red = 0, err_green = 0, err_blue = 0;
	
	for (uint8_t i=0; i<128; i++)
	{
		int16_t newRed = red + err_red;
		int16_t newGreen = green + err_green;
		int16_t newBlue = blue + err_blue;
		
		err_red = newRed & 0x7ff;
		err_green = newGreen & 0x7ff;
		err_blue = (newBlue & 0x7ff);
		
		PALETTE_SetColorRGB (0x780+i, newRed >> 11, newGreen >> 11, newBlue >> 11);
		
		red += d_red;
		green += d_green;
		blue += d_blue;
	}

	// Set the color gradient.
	const uint8_t grad_start = 0;
	uint8_t y;
	for (y=0; y<grad_start; y++)
		ROAD_PATTERN_0[y] = ROAD_SOLID_COLOR(0);
	for (y=grad_start; y<grad_start+127; y++)
		ROAD_PATTERN_0[y] = ROAD_SOLID_COLOR(y-grad_start);
	for (y=grad_start+127; y<224; y++)
		ROAD_PATTERN_0[y] = ROAD_SOLID_COLOR(127);
	
	// Swap buffers.
	ROAD_SwapBuffers ();
}

void SPRITE_SetupStatic (SpriteData* pDest, int16_t left, int16_t top, uint8_t bank, uint16_t offset, int8_t pitch, uint8_t height, uint8_t palette)
{
	pDest->Hidden = pDest->Hidden2 = pDest->EndOfList = 0;
	pDest->EndOfList = 0;
	pDest->Bank = (bank & 3);
	pDest->BankOffset = offset;
	pDest->Top = (top + SPRITE_ORIGIN_TOP);
	pDest->Left = (left + SPRITE_ORIGIN_LEFT);
	pDest->Pitch = pitch;
	pDest->EnableShadows = 0;
	pDest->Priority = 3;
	pDest->HorizontalZoomFactor = pDest->VerticalZoomFactor = 0x200;
	pDest->DrawTopToBottom = pDest->DrawLeftToRight = 1;
	pDest->FlipHorizontal = 1;
	pDest->Height = height - 1;
	pDest->PaletteIdx = palette;
}

// This needs to be done once.
void InitSpritePalette ()
{
	// Sprite palettes.
	static const uint16_t RadioSpritePalette[] = 
	{
		0x000, 0xfff, 0x0f0, 0x00f, 0x000, 0x888, 0x999, 0xaaa, 0xbbb, 0xccc, 0xddd, 0x0f0, 0x0f0, 0x00f, 0x0f0, 0x000, // Radio.
		0x000, 0xfff, 0x000, 0x888, 0x999, 0xaaa, 0xbbb, 0xccc, 0xddd, 0xeee, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, // Knob
		0x000, 0xdd4, 0x882, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, // Display (1: digits, 2: point)
		0x000, 0x0f0, 0x0ff, 0x00f, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0xfff, // Vu meter (1: top left, 2: top left+1, ..., 14: bottom right)
	};
	
	for (uint16_t i=0; i<sizeof(RadioSpritePalette)/sizeof(uint16_t); i++)
		PALETTE_SetColor16 (PALETTE_SPRITE_START_IDX+i, RadioSpritePalette[i]);
}

void UpdateVUPalette (uint8_t left, uint8_t right) // left/Right 0..6
{
	const uint16_t colors[] = { 0x0f0, 0x0f0, 0x0f0, 0x0ff, 0x0ff, 0x00f, 0x00f };
	uint16_t palStartLeft = PALETTE_SPRITE_START_IDX + 3*16 + 1;
	for (uint16_t i = 0; i<left; i++)
		PALETTE_SetColor16 (palStartLeft+i, colors[i]);
	for (uint16_t i = left; i<7; i++)
		PALETTE_SetColor16 (palStartLeft+i, 0);

	uint16_t palStartRight = PALETTE_SPRITE_START_IDX + 3*16 + 8;
	for (uint16_t i = 0; i<right; i++)
		PALETTE_SetColor16 (palStartRight+i, colors[i]);
	for (uint16_t i = right; i<7; i++)
		PALETTE_SetColor16 (palStartRight+i, 0);
}

// Just keep a sprite list in memory which we can modify and then (partially) copy.
#define MAX_SPRITES 8
SpriteData g_spriteList[MAX_SPRITES+1]; // Leave room for terminator.
uint16_t   g_numSprites = 0;

// Can't use memcpy on volatile.
void SPRITE_Copy (volatile SpriteData* dest, const SpriteData* src)
{
	for (uint16_t i=0; i<4; i++)
		dest->RawData[i] = src->RawData[i];
}

// Set this after the last visible entry.
void SPRITE_SetEndOfList (volatile SpriteData* dest)
{
	dest->Hidden = dest->Hidden2 = dest->EndOfList = 1;
}

const uint16_t TuneKnobOffsets[] = { 0x4185, 0x41e9, 0x424d };
static uint16_t s_tuneKnobPos = 1;
const uint16_t FreqOffsets[] = { 0x4131, 0x414d, 0x4169 }; // 69.2, 74.6, 80.3

// This should be done twice, if we want to be able to animate.
void InitRadioSprites ()
{
	SPRITE_SetupStatic (&g_spriteList[g_numSprites++], 122, 166, 3, 0x42c5, 17, 30, 0);                          // Radio background.
	SPRITE_SetupStatic (&g_spriteList[g_numSprites++], 214, 169, 3, TuneKnobOffsets[s_tuneKnobPos],  4, 25, 1);  // Tune knob.
	SPRITE_SetupStatic (&g_spriteList[g_numSprites++], 140, 173, 3, FreqOffsets[s_tuneKnobPos], 4, 7, 2);        // Digital display.
	SPRITE_SetupStatic (&g_spriteList[g_numSprites++], 150, 187, 3, 0x42b1, 4, 5, 3);                            // VU meter. Animate via palette 3.

	// Add more sprites here.

	SPRITE_SetEndOfList (&g_spriteList[g_numSprites]);

	// Blit the list two times.
	for (uint16_t i=0; i<1; i++)
	{
		IRQ4_Wait ();

		for (uint16_t iSprite=0; iSprite<=g_numSprites; iSprite++)
			SPRITE_Copy (SpriteList+iSprite, g_spriteList+iSprite);

		IRQ4_Wait ();
		SPRITE_SwapBuffers ();
	}

	IRQ4_Wait ();
}

bool s_doSpriteUpdate = false; // Flag this to process later on.

void UpdateRadioSprites ()
{
	g_spriteList[1].BankOffset = TuneKnobOffsets[s_tuneKnobPos];
	g_spriteList[2].BankOffset = FreqOffsets[s_tuneKnobPos];
	s_doSpriteUpdate = true;
}

// This is slightly inefficient still.
void BlitUpdatedSprites ()
{
	if (!s_doSpriteUpdate)
		return;
	//IRQ4_Wait ();
	for (uint16_t iSprite=0; iSprite<=g_numSprites; iSprite++)
		SPRITE_Copy (SpriteList+iSprite, g_spriteList+iSprite);
	//IRQ4_Wait ();
	SPRITE_SwapBuffers ();
	//IRQ4_Wait ();
	s_doSpriteUpdate = false;
}

void SetupBackgroundGraphics ()
{
	// Palette used for text (title, menu, hint).
	TEXT_InitDefaultPalette ();

	// Palette gradient (drawn by the road hardware)
	SetupGradient ();
	
	// Setup background graphic.
	UnpackTileGraphics (1, &RadioGraphics);

	// Enable row scroll, for the scrolling clouds, and duplicate the page horizontally.
	TileRegisters.BackgroundMapPages.Pages = 0x0011;
	TileRegisters.BackgroundScrollX.RowScrollEnable = 1; 
	
	// Offset the row scroll values so that the picture aligns with the screen (the background picture is in the bottom left corner in this case).
	for (uint8_t y=0; y<TILE_PAGE_VISIBLE_ROWS; y++)
		TileRegisters.BackgroundRows[y].HorizontalScroll = 24 * 8;
	
	// Radio sprites.
	InitSpritePalette ();
	InitRadioSprites ();

	// Title text.	
	TEXT_GotoXY (14,1);
	TEXT_SetColor (TEXT_Yellow);
	TEXT_Write ("audio sample");
	
	// Instructions. Left/right to select, Accelerate to play.
	TEXT_GotoXY (7,26);
	TEXT_SetColor (TEXT_White);
	TEXT_WriteRawChar (TEXTGLYPH_ARROWLEFT);
	TEXT_Write (" \006or\010 ");
	TEXT_WriteRawChar (TEXTGLYPH_ARROWRIGHT);
	TEXT_Write ("\006 to choose  \010");
	TEXT_WriteRawChar (TEXTGLYPH_ARROWUP);
	TEXT_Write ("\006 to play");
}

// Making this global so that we can detect key repeat (or lack of) inside the callback.
static InputState s_inputState;

#if HAVE_MUTE_OPTIONS

void DrawMute (uint8_t idx, bool muted)
{
	TEXT_GotoXY (33, 21+(idx << 1));
	TEXT_SetColor (TEXT_White);
	TEXTGLYPH gl = muted ? 9 : 11;
	TEXT_WriteRawChar (gl++);
	TEXT_WriteRawChar (gl++);
}

static uint8_t outputBits = DIGITAL_OUT_MUTE_EXTERNAL|DIGITAL_OUT_MUTE_INTERNAL; // Default, inverted (high = mute off).

void MuteBit (InputButton button, void* param)
{
	if (!INPUT_IsButtonPressedNoRepeat (&s_inputState, button))
		return; // Don't repeat these keys.

	uint8_t mask = (uint32_t)param;
	outputBits ^= mask;
	IO_WriteDigital (outputBits);
	DrawMute (mask == DIGITAL_OUT_MUTE_INTERNAL ? 0 : 1, !(outputBits & mask));
}

#endif // HAVE_MUTE_OPTIONS

void InitializeSound ()
{
	// Let go of the Z80 reset.
	PPI_EnableSound (true);

	// Initialize our service routine and set it up at irq 2.
	orsound_init ();
	IRQ2_SetHandler (orsound_update, NULL); // Needs to run first, or we'll lose the previous handler!
	orsound_enable(1);
	
	// Output (disable mute).
	// Bit set = mute disabled = audio enabled.
	IO_WriteDigital (DIGITAL_OUT_MUTE_EXTERNAL|DIGITAL_OUT_MUTE_INTERNAL);
}

MenuState menuState;

// Some metadata about the songs to animate the VU meter.
typedef struct
{
	uint16_t playTime; // Time in frames; 0xffff = infinite.
	uint8_t  vol;      // 0..7, but clamped.
} TrackInfo;
static const TrackInfo s_musicInfo[] =
{
	{ 0xffff, 5 }, // Magical Sound Shower
	{ 0xffff, 5 }, // Passing Breeze
	{ 0xffff, 7 }, // Splash Wave
	{ 0x1300, 3 }, // Last Wave
	{ 0xffff, 2 }, // Ocean Waves
	{ 0x29,   5 }, // Get Ready
	{ 0x48,   5 }, // Congratulation
	{ 0x2d,   5 }, // Check poi
	{ 0xffff, 9 }  // Applause
};

uint32_t s_playTimeStart; // IRQ4 counter when PlaySound was called.
bool isPlaying = false;
static const TrackInfo* s_pCurrentSong = NULL; // Points to track info of current song.

void PlaySound (InputButton button, void* param)
{
	if (button == Button_Accelerate)
	{
		int16_t index = (int16_t)(int)param;
		orsound_write_command (0x93); // NEW_COMMAND
		orsound_write_command (0xA2); // YM_SET_LEVELS.
		orsound_write_command (index); // Waves.
		MENU_SetHighlightItem (&menuState, MENU_GetSelectedItem (&menuState));
		s_playTimeStart = IRQ4_GetCounter();
		s_pCurrentSong = &s_musicInfo[MENU_GetSelectedItem (&menuState)];
		isPlaying = true;
	}
}

static const MenuItem s_menuItems[] = 
{
	{ "Magical Sound Shower", PlaySound, (void*)ORSoundCmd_MagicalSoundShower },
	{ "Passing Breeze",       PlaySound, (void*)ORSoundCmd_PassingBreeze },
	{ "Splash Wave",          PlaySound, (void*)ORSoundCmd_SplashWave },
	{ "Last Wave",            PlaySound, (void*)ORSoundCmd_LastWave },
	{ "Ocean Waves",          PlaySound, (void*)ORSoundCmd_Waves },
	{ "\"Get Ready\"",        PlaySound, (void*)ORSoundCmd_GetReady },
	{ "\"Congratulation\"",   PlaySound, (void*)ORSoundCmd_Congratulations },
	{ "\"Check Poi\"",        PlaySound, (void*)ORSoundCmd_Checkpoint },
	{ "Applause" ,            PlaySound, (void*)ORSoundCmd_Applause },
#if HAVE_MUTE_OPTIONS
	{ "Internal Mute",        MuteBit,   (void*)DIGITAL_OUT_MUTE_INTERNAL },
	{ "External Mute",        MuteBit,   (void*)DIGITAL_OUT_MUTE_EXTERNAL },
#endif // HAVE_MUTE_OPTIONS
};

// Entry point.
void main ()
{
	// Clean init.
	HW_Init (HWINIT_Default, PACKED_COLOR(0x4f90));

	// Initialize sound hardware and service interrupt, does not play anything yet.
	InitializeSound ();
	
	// Set up background graphics and title/hints.	
	SetupBackgroundGraphics ();
	
	// Initialize input helper.
	INPUT_Init (&s_inputState, true); // Run after InitializeSound!

	// Initialize menu, this draws the initial menu state.
	MenuInitParams initParams;
	initParams.iLeft = 8; initParams.iTop = 3;
	initParams.iHighlightItem = 4;
	MENU_Init (&menuState, s_menuItems, sizeof(s_menuItems)/sizeof(MenuItem), &initParams, MIF_POSITION|MIF_HIGHLIGHT);

#if HAVE_MUTE_OPTIONS	
	// Draw 'off' for both mute options.
	DrawMute (0, false);
	DrawMute (1, false);
#endif // HAVE_MUTE_OPTIONS

	// Start 'waves' sound initially.
	// Fudge the selection so that the callback sets up the VU meter correctly for this track.
	menuState.iSelection = 4;
	PlaySound (Button_Accelerate, (void*)ORSoundCmd_Waves);
	menuState.iSelection = 0;
	
	for (;;)
	{
		// Wait for a new vertical blank.
		IRQ4_Wait();

		// Update VU palette, but do this inside the vblank.
		static uint8_t vuDelay = 0;
		if (++vuDelay == 3)
		{
			uint8_t maxVol = 2;
			if (isPlaying && s_pCurrentSong)
			{
				if (s_pCurrentSong->playTime != 0xffff)
				{
					uint16_t playTime = IRQ4_GetCounter()-s_playTimeStart;
					if (playTime >= s_pCurrentSong->playTime)
					{
						// Song ended.
						isPlaying = false;
						s_pCurrentSong = NULL;
					}
				}

				if (isPlaying)
				{
					maxVol = s_pCurrentSong->vol + 1;
				}
			}
			uint8_t vuLeft = rand()%maxVol;
			uint8_t vuRight = rand()%maxVol;
			UpdateVUPalette (vuLeft, vuRight);

			vuDelay = 0;
		} // VU update

#if DEBUG_PLAY_TIME
		// Debug play time here.
		if (isPlaying)
		{
			TEXT_GotoXY (1,2);
			TEXT_SetColor (TEXT_Purple);
			TEXT_WriteHex32 (IRQ4_GetCounter()-s_playTimeStart);
		}
#endif // DEBUG_PLAY_TIME

		INPUT_Update (&s_inputState);

		// Update knob/freq if we move from item to item.
		bool left = INPUT_IsButtonPressed (&s_inputState, Button_Left), right = INPUT_IsButtonPressed (&s_inputState, Button_Right);
		if (left && !right)
		{
			if (s_tuneKnobPos-- == 0)
				s_tuneKnobPos = 2;
			UpdateRadioSprites ();
		}
		else if (right && !left)
		{
			if (++s_tuneKnobPos == 3)
				s_tuneKnobPos = 0;
			UpdateRadioSprites ();
		}

		// Handle any sprites that need to be copied to the hardware sprite list.
		BlitUpdatedSprites();

#if DEBUG_PLAY_TIME
		if (INPUT_IsButtonPressed (&s_inputState, Button_Brake)) // use this to stop the timer.
			isPlaying = false; // stop timer.
#endif // DEBUG_PLAY_TIME

		// Poll input.
		MENU_Update (&menuState, &s_inputState);

		// Scroll clouds in the background.		
		static uint16_t cloudScroll = 0;
		++cloudScroll;
		TileRegisters.BackgroundRows[14].HorizontalScroll = (cloudScroll >> 4);
	}
}
