#include <io.h>
#include <irq.h>
#include <stddef.h>
#include "adchandler.h"

#if !ADC_SUPPORT_ADIR

static uint8_t s_analogInputValues[MAX_ANALOG_INPUTS] = { 0x80, 0x00, /* ... */ }; // Center steering.
static uint8_t s_lastAnalogInputSampleStarted = (uint8_t)-1;
static uint8_t s_numAnalogInputs = MAX_ANALOG_INPUTS_ON_IRQ; // By default, we'll poll the 3 connected ones.

// Limits the number of analog inputs that we update on interrupt.
void ADC_SetMaxAnalogInputs (uint8_t numInputs)
{
	if (numInputs >= MAX_ANALOG_INPUTS)
		return; // Invalid param.
	
	s_numAnalogInputs = numInputs;
}

// Previous IRQ2 function, if any.
static IRQFUNC* s_pOldIrq2Handler = (IRQFUNC*)NULL;

// IRQ handler. Reads the output of the last sample that was triggered, and starts a new one.
static void ADC_AnalogInputHandler ()
{
	if (s_lastAnalogInputSampleStarted != (uint8_t)-1)
	{
		// ADC sample is supposedly ready. Copy it.
		s_analogInputValues[s_lastAnalogInputSampleStarted] = INPUT_ReadAnalogInputImmediate (); // No waiting.
	}
	
	// Start a new sample.
	s_lastAnalogInputSampleStarted++; 
	if (s_lastAnalogInputSampleStarted >= s_numAnalogInputs)
		s_lastAnalogInputSampleStarted = 0;
	
	INPUT_TriggerAnalogInputSample ((AnalogInputPort)s_lastAnalogInputSampleStarted);

	if (s_pOldIrq2Handler)
		s_pOldIrq2Handler ();
}

// Reads buffered analog input.
uint8_t ADC_ReadInputDeferred (uint8_t idx)
{
	if (idx >= s_numAnalogInputs)
		return 0; // Out of range.

	return s_analogInputValues[idx];
}

void ADC_InstallHandler ()
{
	IRQ2_SetHandler (&ADC_AnalogInputHandler, &s_pOldIrq2Handler);
}

#endif // !ADC_SUPPORT_ADIR

uint8_t INPUT_GetAnalogInput (uint8_t idx)
{
#if ADC_SUPPORT_ADIR
	// Trigger a sample and wait for it.
	INPUT_TriggerAnalogInputSample (idx);
	return INPUT_ReadAnalogInputWait ();
#else
	// Read from the buffered table.
	return ADC_ReadInputDeferred (idx);
#endif // !ADC_SUPPORT_ADIR
}
