// OutRun LPT(-ish) background communication layer.
// Quasi hybrid compatibility ("Centronics") + nibble mode.

// LPT/IEEE1284 info:
// http://www.theiling.de/parport/
// http://www.piclist.com/techref/io/parallel/1284/1284int.htm

#include "irq.h"
#include "io.h"
#include "comm.h"
#include <stdlib.h> // NULL

// Enable to visualize time spent via palette (doesn't work in MAME)
#define PROFILE_COMM 1

#if PROFILE_COMM
#include "palette.h"
#include "text.h"
#define PROFILE_COMM_COLOR (PALETTE_RAM_BASE[PALETTE_ROAD_SOLID_START_IDX])

// State-colors
static const uint16_t STATE_COLORS[] = {
	0x444, // kState_Idle,
	0x446, // kState_Ready,
	0x464, // kState_DataAvail,
	0x466, // kState_Ready_DataAvail,

	0x181, // kState_StrobeWait,
	0x228, // kState_NibbleWait1,
	0x248, // kState_NibbleWait2,
	0x228, // kState_NibbleWait3,
};
#endif // PROFILE_COMM

//------------------------------------------------------------------------------------------------------------------------------------------------

// Input (Outrun PCB/Printer <- Host)
// The data bits (D0..D7) are in DIGITAL_INPUT_2.
// INIT and SELECT aren't hooked up on the old (LPT) interface.
enum __attribute__ ((__packed__))
{
	DIG_IN_nAUTOFEED = 1 << 1, // DIGITAL_INPUT_1_Test, // HostBusy.
	DIG_IN_HOSTBUSY  = DIG_IN_nAUTOFEED,
	DIG_IN_nSTROBE   = 1 << 0,
};

// Output (Outrun PCB/Printer -> Host)
// All 5 are hooked up.
enum __attribute__ ((__packed__))
{
	DIG_OUT_nERROR      = 0x01, // D0, D4, DataAvail
	DIG_OUT_SELECT      = 0x02, // D1, D5 (not to be confused with the nSELECT input)
	DIG_OUT_PAPER_END   = 0x04, // D2, D6
	DIG_OUT_BUSY        = 0x08, // D3, D7 (yellow)
	DIG_OUT_NIBBLE_MASK = 0x0f,
	DIG_OUT_nACK        = 0x10, // PtrClk.
};

//------------------------------------------------------------------------------------------------------------------------------------------------

//
// Ringbuffer implementation.
//

void RingBuffer_Clear (RING_BUFFER_INFO* pInfo)
{
	pInfo->ReadPos  = 0;
	pInfo->WritePos = 0;
	pInfo->Usage    = 0;
}

bool RingBuffer_PushBack (RING_BUFFER_INFO* pInfo, uint8_t data)
{
	if (pInfo->Usage == pInfo->Size)
		return false;
	pInfo->Buffer[pInfo->WritePos] = data;
	pInfo->Usage++;
	if ((++pInfo->WritePos) == pInfo->Size)
		pInfo->WritePos = 0;
	return true;
}

inline bool RingBuffer_IsEmpty (const RING_BUFFER_INFO* pInfo) { return pInfo->Usage == 0; }
inline bool RingBuffer_IsFull (const RING_BUFFER_INFO* pInfo) { return pInfo->Usage == pInfo->Size; }

bool RingBuffer_PopFront (uint8_t* pData, RING_BUFFER_INFO* pInfo)
{
	if (pInfo->Usage == 0)
		return false;
	*pData = pInfo->Buffer[pInfo->ReadPos];
	pInfo->Usage--;
	if ((++pInfo->ReadPos) == pInfo->Size)
		pInfo->ReadPos = 0;
	return true;
}

bool RingBuffer_PeekFront (uint8_t* pData, RING_BUFFER_INFO* pInfo)
{
	if (pInfo->Usage == 0)
		return false;
	*pData = pInfo->Buffer[pInfo->ReadPos];
	return true;
}

//==================================================================================================================================

// Used to disable IRQ while inside user functions.
static volatile uint8_t s_commDispatchLock = 0;
static COMM_CONTEXT* s_pCommContext = NULL;
static void _COMM_IRQ_Handler ();

void COMM_Init (COMM_CONTEXT* pContext, uint8_t initDigitalState, bool enableReceive)
{
	// Output buffer.
	pContext->OutBufferInfo.Buffer = pContext->OutBuffer;
	pContext->OutBufferInfo.Size = sizeof(pContext->OutBuffer);
	RingBuffer_Clear (&pContext->OutBufferInfo);

	// Input buffer.
	pContext->InBufferInfo.Buffer = pContext->InBuffer;
	pContext->InBufferInfo.Size = sizeof(pContext->InBuffer);
	RingBuffer_Clear (&pContext->InBufferInfo);

	// Set up initial state.
	pContext->ReceiveEnabled = enableReceive;
	pContext->DispatchState.DataAvailable = false;
	pContext->DispatchState.State = enableReceive ? kState_Ready : kState_Idle;

	// Depending on whether we're immediately going to start receiving, set BUSY flag or not.
	initDigitalState &= ~(DIG_OUT_NIBBLE_MASK);
	initDigitalState |= DIG_OUT_nACK | DIG_OUT_nERROR;
	if (!enableReceive)
		initDigitalState |= DIG_OUT_BUSY;
	pContext->DigitalOutShadow = initDigitalState;
	IO_WriteDigital (pContext->DigitalOutShadow);

	// Set up background dispatch interrupt (irq2)
	s_pCommContext = pContext;
	IRQ2_SetHandler (&_COMM_IRQ_Handler, &pContext->PrevIRQHandler);
}

// Macro used to trigger colors on state change.
#if PROFILE_COMM
#define OUTPUT_STATE(x) { \
	TEXT_RAM_BASE[(TEXT_SCREEN_HEIGHT*TEXT_SCREEN_WIDTH)-3] = ((x+'0') | (TEXT_Red << 9) | 0x8000); \
}
#define CHANGE_STATE(x) { \
	pContext->DispatchState.State = x; \
	PROFILE_COMM_COLOR = STATE_COLORS[x]; \
	OUTPUT_STATE(x); \
}
#else
#define OUTPUT_STATE(x) do{}while(0)
#define CHANGE_STATE(x) { pContext->DispatchState.State = x; }
#endif

void COMM_EnableReceive (COMM_CONTEXT* pContext, bool enableReceive)
{
	if (pContext->ReceiveEnabled == enableReceive)
		return;

	asm ("add.b #1, s_commDispatchLock");

	pContext->ReceiveEnabled = enableReceive;

	struct COMM_STATE* pDS = &pContext->DispatchState;
	switch (pDS->State)
	{
	case kState_Idle:
	case kState_DataAvail:
		if (enableReceive && !RingBuffer_IsFull (&pContext->InBufferInfo))
		{
			pDS->State = (pDS->State == kState_Idle) ? kState_Ready : kState_Ready_DataAvail;
			pContext->DigitalOutShadow &= ~DIG_OUT_BUSY;
			IO_WriteDigital (pContext->DigitalOutShadow);
			OUTPUT_STATE(pDS->State);
		}
		break;

	case kState_Ready:
	case kState_Ready_DataAvail:
		if (!enableReceive)
		{
			pDS->State = (pDS->State == kState_Ready) ? kState_Idle : kState_DataAvail;
			pContext->DigitalOutShadow &= ~DIG_OUT_BUSY;
			IO_WriteDigital (pContext->DigitalOutShadow);
			OUTPUT_STATE(pDS->State);
		}
		break;

	case kState_StrobeWait:
		{
			// Yeah.. about that.
			// Wait for the current strobe to end; we've already buffered the byte anyway.
		}
		break;

	default: // Leave for later.
		break;
	}

	asm ("add.b #-1, s_commDispatchLock");
}

// Returns the number of bytes available for immediate receive.
// Please note that due to locking, hammering this function will slow down or even block transmissions.
uint8_t COMM_ReceiveAvailable (COMM_CONTEXT* pContext)
{
	asm ("add.b #1, s_commDispatchLock");

	uint8_t numAvailable = pContext->InBufferInfo.Usage;

	asm ("add.b #-1, s_commDispatchLock");
	return numAvailable;
}

// Queries the number of buffered and free outgoing bytes.
// Fixme: rename to Get*Info? Doesn't really have a 'state' per se.
void COMM_GetSendBufferState (COMM_CONTEXT* pContext, uint8_t* pSendBuffered, uint8_t* pSendFree)
{
	asm ("add.b #1, s_commDispatchLock");

	*pSendBuffered = pContext->OutBufferInfo.Usage;
	*pSendFree = pContext->OutBufferInfo.Size - pContext->OutBufferInfo.Usage;

	asm ("add.b #-1, s_commDispatchLock");
}

// Non-blocking; will try to buffer as much as possible and return.
uint8_t COMM_Send (COMM_CONTEXT* pContext, const uint8_t* pData, uint8_t len)
{
	if (!len)
		return 0;

	uint8_t remain = len;
	asm ("add.b #1, s_commDispatchLock");

	do
	{
		if (!RingBuffer_PushBack (&pContext->OutBufferInfo, *pData++))
			break; // Full.
	} while (--remain);

	// If in 'ready' mode, set DataAvail (or have the interrupt handle that)
	pContext->DispatchState.DataAvailable = true; // Fixme: we can remove this?

	switch (pContext->DispatchState.State)
	{
	case kState_Idle:
		pContext->DispatchState.State = kState_DataAvail;
		break;

	case kState_Ready:
		pContext->DispatchState.State = kState_Ready_DataAvail;
		break;

	case kState_DataAvail:
	case kState_Ready_DataAvail:
	default:
		// Other states should pick this up later when transitioning back
		// to any of the 4 idle states.
		goto send_done;
	}

	OUTPUT_STATE(pContext->DispatchState.State);

	// Set DataAvail (= nERROR)
	pContext->DigitalOutShadow &= ~DIG_OUT_nERROR;
	IO_WriteDigital (pContext->DigitalOutShadow);

send_done:

	asm ("add.b #-1, s_commDispatchLock");
	return len - remain;
}

uint8_t COMM_Receive (COMM_CONTEXT* pContext, uint8_t* pData, uint8_t len)
{
	if (!len)
		return 0;

	uint8_t remain = len;
	asm ("add.b #1, s_commDispatchLock");

	do
	{
		if (!RingBuffer_PopFront (pData, &pContext->InBufferInfo))
			break; // No more data.
		pData++;
	} while (--remain);

	// The buffer can't be full at this point.
	// Release BUSY if we allow incoming data.
	pContext->ReceiveEnabled = true;

	if (pContext->ReceiveEnabled)
	{
		switch (pContext->DispatchState.State)
		{
		case kState_Idle:
			pContext->DispatchState.State = kState_Ready;
			break;
		case kState_DataAvail:
			pContext->DispatchState.State = kState_Ready_DataAvail;
			break;

		case kState_Ready:
		case kState_Ready_DataAvail:
			// AllowReceive == true, pin already low.
		default:
			// Don't update yet.
			goto receive_done;
		}

		OUTPUT_STATE(pContext->DispatchState.State);

		// Update BUSY pin.
		pContext->DigitalOutShadow &= ~DIG_OUT_BUSY;
		IO_WriteDigital (pContext->DigitalOutShadow);
	}

receive_done:

	asm ("add.b #-1, s_commDispatchLock");
	return len - remain;
}

// Set up nibble data pins.
static bool _COMM_SetupNibble (COMM_CONTEXT* pContext, uint8_t nibble)
{
	// Set data, and wait a bit before strobing.
	pContext->DigitalOutShadow &= ~DIG_OUT_NIBBLE_MASK;
	// PTRCLK should still be high.
	pContext->DigitalOutShadow |= (nibble & DIG_OUT_NIBBLE_MASK);
	IO_WriteDigital (pContext->DigitalOutShadow);

	asm ("nop;"
	     "nop;"
	     "nop;"
	     "nop;"
	     "nop;");

	// nAUTOFEED low = Set PtrClk to low. (Was 0x9x).
	pContext->DigitalOutShadow &= ~DIG_OUT_nACK;
	IO_WriteDigital (pContext->DigitalOutShadow);
}

static bool _COMM_Handle_Send (COMM_CONTEXT* pContext, uint8_t controlIn)
{
	// Check for HostBusy low (nibble mode)
	if ((controlIn & DIG_IN_nAUTOFEED) == 0)
	{
		// HostBusy low.
		// Put up the first half of the nibble and PtrClk.
		uint8_t b;
		RingBuffer_PeekFront (&b, &pContext->OutBufferInfo);
		_COMM_SetupNibble (pContext, (b & 0xf)); // Low, high.

		// Wait for HostBusy high.
		CHANGE_STATE(kState_NibbleWait1);
		return true;
	}

	return false;
}

static bool _COMM_Handle_Receive (COMM_CONTEXT* pContext, uint8_t controlIn)
{
	// Check for nSTROBE low.
	if ((controlIn & DIG_IN_nSTROBE) == 0)
	{
		// Strobe low, data valid.
		uint8_t data = INPUT_ReadDigital (DIGITAL_INPUT_2);
		RingBuffer_PushBack (&pContext->InBufferInfo, data);

		// Go busy again
		// Set nERROR (DataAvail) high as well for good measure, so we don't accidentally
		// come out of kState_StrobeWait mistaking the nACK low for a PtrClk.
		pContext->DigitalOutShadow |= (DIG_OUT_BUSY | DIG_OUT_nERROR);
		IO_WriteDigital (pContext->DigitalOutShadow);

		// Ack
		pContext->DigitalOutShadow &= ~DIG_OUT_nACK;
		IO_WriteDigital (pContext->DigitalOutShadow);

		// Wait for strobe high.
		CHANGE_STATE(kState_StrobeWait);
		return true;
	}

	return false;
}

static bool _COMM_Setup_Initial_State (COMM_CONTEXT* pContext)
{
	// Decide the next state.
	// We just added a character to the buffer, so it's possible that it's now full.
	struct COMM_STATE* pDS = &pContext->DispatchState;
	bool allowReceive = pContext->ReceiveEnabled && !RingBuffer_IsFull (&pContext->InBufferInfo);
	pDS->State = pDS->DataAvailable ?
		(allowReceive ? kState_Ready_DataAvail : kState_DataAvail) :
		(allowReceive ? kState_Ready : kState_Idle);

	uint8_t digShadow = pContext->DigitalOutShadow | DIG_OUT_nERROR | DIG_OUT_BUSY;
	if (allowReceive)
		digShadow &= ~DIG_OUT_BUSY;
	if (pDS->DataAvailable)
		digShadow &= ~DIG_OUT_nERROR;

	if (digShadow != pContext->DigitalOutShadow)
	{
		pContext->DigitalOutShadow = digShadow;
		IO_WriteDigital (digShadow);
	}

	CHANGE_STATE(pDS->State);
}

// Single dispatch. 
// Returns true on state change.
/* static */ bool _COMM_Dispatch_Single (COMM_CONTEXT* pContext)
{
	uint8_t controlIn = INPUT_ReadDigital (DIGITAL_INPUT_1);
	struct COMM_STATE* pDS = &pContext->DispatchState;

	switch (pDS->State)
	{
	case kState_Idle:
		return false; // Nothing to do except wait for user changes.

	case kState_Ready:
		return _COMM_Handle_Receive (pContext, controlIn);

	case kState_DataAvail:
		return _COMM_Handle_Send (pContext, controlIn);

	case kState_Ready_DataAvail:
		return _COMM_Handle_Receive (pContext, controlIn) ||
		       _COMM_Handle_Send (pContext, controlIn);

	case kState_StrobeWait:
		{
			// Wait for nSTROBE to go high again.
			if ((controlIn & DIG_IN_nSTROBE) != 0)
			{
				// Release ack
				pContext->DigitalOutShadow |= DIG_OUT_nACK;
				IO_WriteDigital (pContext->DigitalOutShadow);

				_COMM_Setup_Initial_State (pContext);
				return true;
			}
		}
		break;

	case kState_NibbleWait1: // 5; PtrClk was low in response to HostBusy low.
		{
			// Wait for HostBusy high (nibble ack)
			if ((controlIn & DIG_IN_nAUTOFEED) != 0)
			{
				// HostBusy high again.
				// Set PtrClk high, wait for HostBusy low again.
				pContext->DigitalOutShadow |= DIG_OUT_nACK; //??
				IO_WriteDigital (pContext->DigitalOutShadow);

				CHANGE_STATE(kState_NibbleWait2);
				return true;
			}
		}
		break;

	case kState_NibbleWait2: // 6
		{
			// Wait for HostBusy low.
			if ((controlIn & DIG_IN_nAUTOFEED) == 0)
			{
				// HostBusy low.
				uint8_t b;
				RingBuffer_PopFront (&b, &pContext->OutBufferInfo);

				_COMM_SetupNibble (pContext, (b >> 4)); // Low, high.
				// This also sets PtrClk low.

				// This removed the byte from the buffer, so we may no longer have data available now.
				if (RingBuffer_IsEmpty (&pContext->OutBufferInfo))
					pDS->DataAvailable = false; // Could be moved to NibbleWait3.

				CHANGE_STATE(kState_NibbleWait3);
				return true;
			}
		}
		break;

	case kState_NibbleWait3: // 7
		{
			// PtrClk still low, wait for HostBusy again.
			if ((controlIn & DIG_IN_nAUTOFEED) != 0)
			{
				// HostBusy high.
				// Set DataAvail to the appropriate value.
				if (pDS->DataAvailable)
					pContext->DigitalOutShadow &= ~DIG_OUT_nERROR;
				else
					pContext->DigitalOutShadow |= DIG_OUT_nERROR;
				IO_WriteDigital (pContext->DigitalOutShadow);

				// PtrClk High.
				pContext->DigitalOutShadow |= DIG_OUT_nACK;
				IO_WriteDigital (pContext->DigitalOutShadow);
				_COMM_Setup_Initial_State (pContext);
				return true;
			}
		}
		break;
	}

	return false; // No state change.
}

#undef CHANGE_STATE

// Multiple-dispatch, keeping track of cost.
static bool _COMM_Dispatch (COMM_CONTEXT* pContext)
{
	// Since this could be eating into our frame time considerably, here are some limits to tweak the maximum usage.
	const int16_t kStateCost = 16,            // (average) cost of a state transition. 
	              kWaitCost  = 1,             // cost of a state transition that didn't happen
	              kMaxConsecutiveWait = 20,   // break out early?
	              kMaxReadyWait = 5;          // same as above but for the READY state.

	// 4k/sec, 60 frames, IRQ2 (3x/frame), 2 states for compatibility mode.
	const int16_t kFrameBudget = (2048 / 60 / 3) * 2 * kStateCost;

	bool bAnyStateChange = false;
	int16_t frameCost = 0, numWait = 0;
	do
	{
		bool bStateChanged = _COMM_Dispatch_Single (pContext);
		bAnyStateChange |= bStateChanged;

		frameCost += (bStateChanged ? kStateCost : kWaitCost);

		if (bStateChanged)
		{
#if PROFILE_COMM
			PROFILE_COMM_COLOR ^= 0x282;
#endif // PROFILE_COMM
			numWait = 0;
		}
		else
		{
			numWait++;

			if (pContext->DispatchState.State == kState_Idle)
				break; // This won't do anything unless we get some user calls.
			if (pContext->DispatchState.State == kState_Ready && numWait >= kMaxReadyWait) // Tweak this value.
				break;
		} 
	} while (frameCost <= kFrameBudget && numWait < kMaxConsecutiveWait);

	return bAnyStateChange;
}

static void _COMM_IRQ_Handler ()
{
#if PROFILE_COMM
	uint16_t prevColor = PROFILE_COMM_COLOR;
	PROFILE_COMM_COLOR ^= 0xf0;
#endif // PROFILE_COMM

	// Run previous IRQ handler.
	if (s_pCommContext->PrevIRQHandler)
		s_pCommContext->PrevIRQHandler ();

#if PROFILE_COMM
	PROFILE_COMM_COLOR ^= 0xf0;
#endif // PROFILE_COMM

	if (!s_commDispatchLock)
	{
#if PROFILE_COMM
		PROFILE_COMM_COLOR = STATE_COLORS[s_pCommContext->DispatchState.State];
#endif // PROFILE_COMM

		_COMM_Dispatch (s_pCommContext);
	}

#if PROFILE_COMM
	PROFILE_COMM_COLOR = prevColor;
#endif // PROFILE_COMM
}
