#include "input.h"
#include <io.h>
#include <adchandler.h>
#include <irq.h>

void INPUT_Init (InputState* pInputState, bool initAnalog)
{
	pInputState->ButtonsDown = 0;
	pInputState->ButtonsPressed = 0;
	pInputState->ButtonsChanged = 0;

#if !ADC_SUPPORT_ADIR
	// Install IRQ2 handler.
	if (initAnalog)
		ADC_InstallHandler ();
#endif // !ADC_SUPPORT_ADIR

#if INPUT_USE_IRQ_TIME
	pInputState->LastUpdateTime = (uint16_t)IRQ4_GetCounter();
#endif // INPUT_USE_IRQ_TIME
}

// Call each frame.
void INPUT_Update (InputState* pInputState)
{
	// Read analog axes (blocking or through IRQ handler, based on ADC_SUPPORT_ADIR.
	for (uint8_t i=0; i<MAX_AXES; i++)
		pInputState->AxisValues[i] = INPUT_GetAnalogInput(i);
	
	// Read digital buttons.
	uint16_t buttonState = INPUT_ReadDigital(DIGITAL_INPUT_1) ^ 0xff;

	// Emulate digital buttons.
	if (pInputState->AxisValues[Axis_Steering] < 0x60)
		buttonState |= (1 << Button_Left);
	if (pInputState->AxisValues[Axis_Steering] > 0xa0)
		buttonState |= (1 << Button_Right);
	if (pInputState->AxisValues[Axis_Accelerator] >= 0x40)
		buttonState |= (1 << Button_Accelerate);
	if (pInputState->AxisValues[Axis_Brake] >= 0x40)
		buttonState |= (1 << Button_Brake);
	
	uint16_t buttonsChanged = (pInputState->ButtonsDown ^ buttonState);
	pInputState->ButtonsDown = buttonState;
	pInputState->ButtonsChanged = buttonsChanged;
	
	uint16_t buttonsPressed = 0;
	uint16_t buttonMask = 1;

#if INPUT_USE_IRQ_TIME
	uint16_t curTime =(uint16_t)IRQ4_GetCounter();
	uint16_t timeElapsed = curTime - pInputState->LastUpdateTime;
	if (timeElapsed == 0)
		timeElapsed = 1;
	pInputState->LastUpdateTime = curTime;
#else
	uint16_t timeElapsed = 1;
#endif // !INPUT_USE_IRQ_TIME

	for (uint8_t i=0; i<MAX_BUTTONS; i++)
	{
		if (buttonsChanged & buttonState & buttonMask)
		{
			// This button was just pressed.
			pInputState->TimeOutValues[i] = 15;
			buttonsPressed |= buttonMask;
		}
		else if (buttonState & buttonMask)
		{
			// Button is being held.
			if (timeElapsed >= pInputState->TimeOutValues[i])
			{
				// Triggered repeat.
				pInputState->TimeOutValues[i] = 2;
				buttonsPressed |= buttonMask;
			}
			else
			{
				pInputState->TimeOutValues[i] -= timeElapsed;
			}
		}
		
		buttonMask <<= 1;
	}
	
	pInputState->ButtonsPressed = buttonsPressed;
}
