#include "text.h"
#include "paletteutil.h"
#include <stdbool.h>
#include <stddef.h>
#include <irq.h>

//------------------------------------------------------------------------------------------------------------------------

static TEXT_WINDOW_STATE s_windowState = 
{
	TEXT_SCREEN_WIDTH-TEXT_SCREEN_VISIBLE_WIDTH,
	0,
	TEXT_SCREEN_WIDTH,
	TEXT_SCREEN_HEIGHT,
	0x8000 | (TEXT_White << 9),
	TEXT_SCREEN_WIDTH-TEXT_SCREEN_VISIBLE_WIDTH,
	0,
	true,   // ScrollEnabled
	false,  // CursorEnabled
	true,   // AutoWordWrap
};

//------------------------------------------------------------------------------------------------------------------------

void TEXT_InitDefaultPalette ()
{
	// Sets up 8 different colors sets. We don't have black, because it's ugly.
	uint16_t offset;
	for (offset=1; offset<7; offset++)
	{
		uint16_t gradient = (offset*9)+9;
		uint16_t high = gradient;
		uint16_t low = gradient >> 1;
		if (high > 31)
			high = 31;

		PALETTE_SetColorRGB( 0+offset, high, low,  low) ; // Red
		PALETTE_SetColorRGB( 8+offset, low,  high, low) ; // Green
		PALETTE_SetColorRGB(16+offset, low,  low,  high); // Blue

		PALETTE_SetColorRGB(24+offset, high, high, low ); // Yellow
		PALETTE_SetColorRGB(32+offset, high, low,  high); // Purple
		PALETTE_SetColorRGB(40+offset, low,  high, high); // Cyan
	
		PALETTE_SetColorRGB(48+offset, low,  low,  low ); // Gray
		PALETTE_SetColorRGB(56+offset, high, high, high); // White
	}

	// Background and outline.
	for (offset=0; offset<64; offset+=8)
	{
		PALETTE_SetColor16 (offset,   0); // Background.
		PALETTE_SetColor16 (offset+7, 0); // Outline.
	}
}

// Animating cursor.
static volatile int8_t   s_cursorVisibleCount = 0; // <= 0 = hidden (refcounted); 1 = drawn.
static volatile uint16_t s_cursorSavedChar;        // Character that was under the cursor.
static volatile uint8_t  s_cursorAnimCounter = 0xff;

// Defined as (counter & mask) == 0 -> visible.
#define CURSOR_ANIM_MASK 16
// Todo: could have a dynamic mask based on right edge or not.

static IRQFUNC* s_pPrevIRQ4Handler    = NULL;
static bool     s_irqHandlerInstalled = false;

static void _IRQ_CursorUpdate ()
{
	if (s_pPrevIRQ4Handler)	
		s_pPrevIRQ4Handler ();

	if (s_cursorVisibleCount <= 0)
		return;

	uint8_t prevCounter = s_cursorAnimCounter++;
	if ((prevCounter ^ s_cursorAnimCounter) & CURSOR_ANIM_MASK)
	{
		uint16_t* pDest = TEXT_RAM_BASE + (s_windowState.CursorY * TEXT_SCREEN_WIDTH) + s_windowState.CursorX;
		if (s_windowState.CursorX == s_windowState.WindowRight)
			pDest--;

		if (prevCounter & CURSOR_ANIM_MASK)
		{
			s_cursorSavedChar = *pDest;
			*pDest = 0xf7 + s_windowState.ColorMask;
		}
		else
		{
			*pDest = s_cursorSavedChar;
		}
	}
}

static void _TEXT_SaveCursor ()
{
	asm ("add.b #-1, s_cursorVisibleCount"); // Make sure this is atomic.

	// Definitely no longer 1 now, so the IRQ doesn't do anything. 
	if (s_cursorVisibleCount == 0)
	{
		// Remove the cursor and draw back the saved character.
		if ((s_cursorAnimCounter & CURSOR_ANIM_MASK) == 0)
		{
			// It was visible.
			uint16_t* pDest = TEXT_RAM_BASE + (s_windowState.CursorY * TEXT_SCREEN_WIDTH) + s_windowState.CursorX;
			if (s_windowState.CursorX == s_windowState.WindowRight)
				pDest--;
			*pDest = s_cursorSavedChar;
			s_cursorAnimCounter = 0xff;
		}
	}
}

static void _TEXT_RestoreCursor ()
{
	// s_cursorVisibleCount must be <= 0 at this point.
	if (s_cursorVisibleCount == 0)
	{
		uint16_t* pDest = TEXT_RAM_BASE + (s_windowState.CursorY * TEXT_SCREEN_WIDTH) + s_windowState.CursorX;
		if (s_windowState.CursorX == s_windowState.WindowRight)
			pDest--;

		// Already 0xff.
		s_cursorAnimCounter = 0xff; // next irq will paint.
		
		// .. or we could paint here (starting at 0).
	}

	asm ("add.b #1, s_cursorVisibleCount");
}

void TEXT_SetCursorEnable (bool enableCursor)
{
	if (s_windowState.CursorEnabled == enableCursor)
		return;

	s_windowState.CursorEnabled = enableCursor;
	if (enableCursor)
	{
		s_cursorVisibleCount = 0; // Restore will add 1.

		// Install IRQ if not yet installed.
		if (!s_irqHandlerInstalled)
		{
			IRQ4_SetHandler (&_IRQ_CursorUpdate, &s_pPrevIRQ4Handler);
			s_irqHandlerInstalled = true;
		}

		_TEXT_RestoreCursor (); // s_cursorVisibleCount: 0 -> 1
	}
	else
	{
		_TEXT_SaveCursor (); // s_cursorVisibleCount: 1 -> 0

		// Could uninstall IRQ handler here.
	}
}

//------------------------------------------------------------------------------------------------------------------------

void TEXT_ClearTextRAM ()
{
	asm ("add.b #-1, s_cursorVisibleCount"); // Not a *real* save, but disable blink anyway.

	// Clear 0x000 through 0xDFE with TEXTGLYPH_EMPTY.
	uint16_t offset;
	for (uint16_t offset=0; offset<0x700; offset++)
		TEXT_RAM_BASE[offset] = TEXTGLYPH_EMPTY;

	_TEXT_RestoreCursor ();
}

void TEXT_SetWindow (uint8_t left, uint8_t top, uint8_t right, uint8_t bottom, bool enableScroll)
{
	// Sanity checks.
	if (right <= left || bottom <= top)
		return;

	_TEXT_SaveCursor ();
	if (right > TEXT_SCREEN_WIDTH)
		right = TEXT_SCREEN_WIDTH;
	if (bottom > TEXT_SCREEN_HEIGHT)
		bottom = TEXT_SCREEN_HEIGHT;

	s_windowState.WindowLeft   = left;
	s_windowState.WindowTop    = top;
	s_windowState.WindowRight  = right;
	s_windowState.WindowBottom = bottom;
	
	// Enable/disable scrolling on this viewport.
	s_windowState.ScrollEnabled = enableScroll;

	// Set the cursor at the top left.
	s_windowState.CursorX = left;
	s_windowState.CursorY = top;

	s_cursorVisibleCount = s_windowState.CursorEnabled ? 0 : -1; // Restore will add 1.
	_TEXT_RestoreCursor ();
}

void TEXT_SaveWindowState (TEXT_WINDOW_STATE* pState)
{
	*pState = s_windowState;
}

void TEXT_RestoreWindowState (const TEXT_WINDOW_STATE* pState)
{
	// This should first restore the cursor of any previously active window.
	_TEXT_SaveCursor ();

	// Copy state.
	s_windowState = *pState;

	// Activate cursor, if any.
	s_cursorVisibleCount = s_windowState.CursorEnabled ? 0 : -1; // Restore will add 1.
	_TEXT_RestoreCursor ();
}

void TEXT_SetColor (uint8_t colorSetIdx)
{
	_TEXT_SaveCursor ();
	colorSetIdx &= 0x7;
	s_windowState.ColorMask = (colorSetIdx << 9) | 0x8000; // Give text priority against sprites.
	_TEXT_RestoreCursor (); // 2
}

void TEXT_GotoXY (uint8_t x, uint8_t y)
{
	_TEXT_SaveCursor ();
	s_windowState.CursorX = x + s_windowState.WindowLeft;
	s_windowState.CursorY = y + s_windowState.WindowTop;
	_TEXT_RestoreCursor ();
}

// Retrieves the cursor position, relative to the console window.
void TEXT_GetXY (uint8_t* pX, uint8_t* pY)
{
	if (pX)
		*pX = s_windowState.CursorX - s_windowState.WindowLeft;
	if (pY)
		*pY = s_windowState.CursorY - s_windowState.WindowTop;
}

// Retrieve window size.
void TEXT_GetWindowSize (uint8_t* pWidth, uint8_t* pHeight)
{
	if (pWidth)
		*pWidth = (s_windowState.WindowRight - s_windowState.WindowLeft);
	if (pHeight)
		*pHeight = (s_windowState.WindowBottom - s_windowState.WindowTop);
}

// Clears the remainder of the current line (within the window) from the current X position, while preserving the position.
void TEXT_ClearEOL ()
{
	_TEXT_SaveCursor ();
	uint16_t* pDest = TEXT_RAM_BASE + (s_windowState.CursorY * TEXT_SCREEN_WIDTH) + s_windowState.CursorX;
	for (uint8_t x = s_windowState.CursorX; x < s_windowState.WindowRight; x++)
		*pDest++ = TEXTGLYPH_EMPTY;
	_TEXT_RestoreCursor ();
}

// Note: internal use only; assumes cursor already saved;
// does not adjust Y position.
static void _TEXT_ScrollWindow ()
{
	uint16_t x, y;
	uint16_t* pDest = TEXT_RAM_BASE + (s_windowState.WindowTop * TEXT_SCREEN_WIDTH);
	const uint16_t* pSrc = pDest + TEXT_SCREEN_WIDTH;
	for (y=s_windowState.WindowTop; y<(s_windowState.WindowBottom-1); y++)
	{
		for (x=s_windowState.WindowLeft; x<s_windowState.WindowRight; x++)
			pDest[x] = pSrc[x];
		pDest += TEXT_SCREEN_WIDTH;
		pSrc  += TEXT_SCREEN_WIDTH;
	}

	// Clear the bottom row.
	for (x=s_windowState.WindowLeft; x<s_windowState.WindowRight; x++)
		pDest[x] = TEXTGLYPH_EMPTY;
}

// Wraps at WindowRight (beyond the last visible character)
static bool _TEXT_HandleCursorWrap ()
{
	if (s_windowState.CursorX == s_windowState.WindowRight)
	{
		if (s_windowState.CursorY == (s_windowState.WindowBottom-1) && s_windowState.ScrollEnabled)
		{
			_TEXT_ScrollWindow();
			s_windowState.CursorX = s_windowState.WindowLeft;
			return true;
		}
		else if (s_windowState.CursorY < (s_windowState.WindowBottom-1))
		{
			s_windowState.CursorY++;
			s_windowState.CursorX = s_windowState.WindowLeft;
			return true;
		}

		return false; // No more space.
	}
	else return true; // Not at the end yet.
}

void TEXT_WriteRawCharFast (const TEXTGLYPH c)
{
	// Write out new character.
	TEXT_RAM_BASE[(s_windowState.CursorY*TEXT_SCREEN_WIDTH)+s_windowState.CursorX] = (c & TEXT_GLYPH_MASK) | s_windowState.ColorMask;
	
	// Update screen position.
	s_windowState.CursorX++;
}

// Let's build word wrapping into this function.
void TEXT_WriteRawChar (TEXTGLYPH c)
{
	_TEXT_SaveCursor ();

	c &= TEXT_GLYPH_MASK;

	if (s_windowState.AutoWordWrap)
	{
		uint8_t prevCursorX = s_windowState.CursorX, 
		        prevCursorY = s_windowState.CursorY;

		if (!_TEXT_HandleCursorWrap())
		{
			_TEXT_RestoreCursor ();
			return; // Doesn't fit on the screen.
		}

		// If we just moved to a new line, it's possible that we want to take the last word with us.
		if (prevCursorX == s_windowState.WindowRight)
		{
			if (c == TEXTGLYPH_EMPTY)
			{
				_TEXT_RestoreCursor ();
				return; // Wrapped due to single space; don't print.
			}

			uint8_t windowHeight = s_windowState.WindowBottom - s_windowState.WindowTop;
			if (windowHeight > 1) // We can only move the word from the previous line if there is a visible previous line.
			{
				// Compensate for scrolling up (needed for word wrap)
				if (s_windowState.CursorY == prevCursorY)
					prevCursorY--; // We scrolled.			

				// We're at the right end. Reverse scan to find a word break.
				TEXTGLYPH* p = TEXT_RAM_BASE + (prevCursorY * TEXT_SCREEN_WIDTH);
				
				// Find the start of the last word.
				uint8_t wordStartX = s_windowState.WindowLeft; 
				for (uint8_t x = prevCursorX-1; x>s_windowState.WindowLeft; x--)
				{
					if ((p[x] & TEXT_GLYPH_MASK) == TEXTGLYPH_EMPTY)
					{
//						p[x] = 0xfe3b; // Draw '->' markers at wrap points.
						wordStartX = x + 1;
						break;
					}
				}

				if (wordStartX > s_windowState.WindowLeft)
				{
					TEXTGLYPH* src = TEXT_RAM_BASE + (prevCursorY*TEXT_SCREEN_WIDTH) + wordStartX;
					TEXTGLYPH* dest = TEXT_RAM_BASE + (s_windowState.CursorY*TEXT_SCREEN_WIDTH) + s_windowState.CursorX;
					uint8_t wordLen = s_windowState.WindowRight-wordStartX;
					for (uint8_t i=0; i<wordLen; i++)
					{
						TEXTGLYPH t = src[i];
						src[i] = (t & ~TEXT_GLYPH_MASK) | TEXTGLYPH_EMPTY; // Replace with space.
						dest[i] = t;
					}
					s_windowState.CursorX += wordLen;
				}
			} // WindowHeight > 1
		} // Right edge
	} // AutoWordWrap

	// Write out new character.
	TEXT_RAM_BASE[(s_windowState.CursorY*TEXT_SCREEN_WIDTH)+s_windowState.CursorX] = c | s_windowState.ColorMask;
	s_windowState.CursorX++;
	
	// Update scrolling for non-word-wrap here, we just immediately switch to the next line.
	if (!s_windowState.AutoWordWrap)
	{
		if (!_TEXT_HandleCursorWrap ()) // Note: this can overflow if scrolling is disabled.
			s_windowState.CursorX--;    // So just go back to the previous position then.
	}

	_TEXT_RestoreCursor ();
}

// Does not support '\n' or colors.
// Does support word wrap/scroll/windowing
void TEXT_WriteRaw (const TEXTGLYPH* glyphs, uint16_t num) 
{
	_TEXT_SaveCursor ();
	while (num--)
		TEXT_WriteRawChar (*glyphs++);
	_TEXT_RestoreCursor ();
}

// Does not support word wrap/scroll/windowing
void TEXT_WriteRawFast (const TEXTGLYPH* glyphs, uint16_t num)
{
	while (num--)
		TEXT_WriteRawCharFast (*glyphs++);
}

// Convert ASCII character to closest glyph in the original tile ROM.
TEXTGLYPH TEXT_GlyphFromASCII (char c)
{
	switch (c)
	{
	case 'A'...'Z':
	case '0'...'9':
		return c;
		
	case 'a'...'z':
		return c = 'A' + (c-'a');
	
	case ' ':
	case '!':
	case '#':
	case '%':
	case '&':
	case '*':
	case '+':
	case '-':
		return (TEXTGLYPH)c;

	case '=':
		return TEXTGLYPH_EQUAL;
	case '.':
		return TEXTGLYPH_PERIOD;
	case '/':
		return TEXTGLYPH_SLASH;
	case '\'':
		return TEXTGLYPH_SINGLEQUOTE;
	case '\"':
		return TEXTGLYPH_DOUBLEQUOTE;
	case '_':
		return 0xe; // Part of an outline, which paints the lower 2 pixels over the entire width.
		
	default:
		return TEXTGLYPH_CROSS;
	}
}

void TEXT_WriteChar (char c)
{
	_TEXT_SaveCursor ();
	switch (c)
	{
	case '\n':
		{
			s_windowState.CursorX = s_windowState.WindowLeft;
			s_windowState.CursorY++;
			if (s_windowState.CursorY == s_windowState.WindowBottom)
			{
				if (s_windowState.ScrollEnabled)
					_TEXT_ScrollWindow();
				s_windowState.CursorY--;
			}
		}
		break;
	
	case '\0': // Fixme: is this used?
		break;
		
	default:
		{
			TEXTGLYPH glyph = TEXT_GlyphFromASCII(c);
			TEXT_WriteRawChar (glyph);
		}
	}
	_TEXT_RestoreCursor ();
}

void TEXT_Write (const char* text)
{
	_TEXT_SaveCursor ();

	const char* textPtr = text;
	char c;
	while ((c = *textPtr++))
	{
		if (c <= 8)
			TEXT_SetColor(c-1);
		else
			TEXT_WriteChar (c);
	}

	_TEXT_RestoreCursor ();
}

const char* TEXT_CalcRect (uint16_t* pWidth, uint16_t* pHeight, const char* pText, bool bWrap)
{
	const uint16_t maxWidth = *pWidth;
	const uint16_t maxHeight = *pHeight;
	uint16_t minWidth = 0; // Largest line width.
	uint16_t x = 0, y = 0;
	
	if (!maxWidth || !maxHeight)
		return NULL; // Can't fit jack shit.
	
	if (!bWrap)
	{
		bool done = false;
		while (!done)
		{
			char c = *pText;
			switch (c)
			{
			case 0:
				done = true;
				break; // All done.
			case 1 ... 8:
				++pText; // Skip over color markup.
				break;
			case '\n':
				x=0; ++y; // Newline.
				if (y == maxHeight)
					done = true; // Filled up our entire window.
				else
					++pText;
				break;
				
			default:
				// Got a text character here.
				// Assume we're still inside our window.
				++x; // So we can do this.
				if (x > minWidth)
					minWidth = x;
				
				if (x == maxWidth) // If we're at the end, we shall now move on to a new line.
				{
					x = 0;
					++y;
					if (y == maxHeight)
						done = true;
				}
				
				if (!done)
					++pText;
			} // switch (c)
		} // for (;;)
	}
	else
	{
		// Keep walking the current word, and then fit it to the screen.
		const char* curStart = pText;
		const char* curEnd = pText;
		const uint16_t windowWidth = maxWidth;
		uint16_t wordLength = 0;
		bool done = false;
	
		while (!done)
		{
			// Inspect the byte at our current position.
			char c = *curEnd;
			switch (c)
			{
			case 1 ... 8:
				// If there's no word buffered, process colors immediatly.
				// Otherwise they're part of the word.
				if (!wordLength)
					curStart = ++curEnd;
				else 
					++curEnd; // Add it to the word (but don't count as length).
				break;
				
			case '\0':
			case ' ':
			case '\n':
				{
					bool stealWhitespace = false;
					if (curEnd > curStart)
					{
						// There was a buffered word. Fit it to the current line.
						uint16_t lineRemaining = windowWidth - x;
						if (wordLength <= lineRemaining)
						{
							// Special case: if we got onto a new line PRECISELY, ignore the space or newline.
							stealWhitespace = (wordLength == lineRemaining);
						}
						else if (wordLength <= windowWidth)
						{
							// Can fit on the next line.
							x=0; ++y;
							if (y == maxHeight)
							{
								// The new word won't fit. Screen is full.
								// pHeight is correct.
								*pWidth = minWidth;
								return curStart;
							}
						}
						else
						{
							// It won't fit on the new line either. Just proceed as if we had no wrapping (also: do not wrap for the start).
						}
						
						// Write the buffered word.
						while (curStart != curEnd)
						{
							char c = *curStart++;
							if (c > 8)
							{
								// This would yield a character. Handle newlines and the likes here.
								++x;
								if (x > minWidth)
									minWidth = x;
								
								if (x == maxWidth)
								{
									// Move on to new line, if we can.
									x=0;
									++y;
									if (y == maxHeight)
									{
										// Wrapped over the edge on the last line.
										// pHeight is correct.
										*pWidth = minWidth;
										return curStart;
									}
								}
							}
						}
					}
						
					// Done?
					if (!c)
					{
						pText = curEnd; // For the return value.
						done = true;
						break;
					}
					
					if (!stealWhitespace)
					{
						if (c == '\n')
						{
							x=0; ++y;
							if (y == maxHeight)
							{
								// Wrapped over the edge on the last line.
								// pHeight is correct.
								*pWidth = minWidth;
								return curEnd;
							}
						}
						else
						{
							++x;
							if (x > minWidth && !wordLength) // Don't count spaces as last character.
								minWidth = x;
							
							if (x == maxWidth)
							{
								x=0;
								++y;
								if (y == maxHeight)
								{
									*pWidth = minWidth;
									return curEnd;
								}
							}
						}
					}
					
					// Advance.
					curStart = ++curEnd;
					wordLength = 0;
				}
				break;
				
			default:
				// Add to the current word.
				++curEnd;
				++wordLength;
				break;
			}
		}	
	}

	// Copy output values.
	*pWidth = minWidth;
	
	// Note: This won't match the TEXT_Write function exactly, because it'd scroll regardless of whether something comes next.
	// Best to disable scrolling when drawing measured text (if it fits onto the screen).
	if (x > 0)
		*pHeight = y+1; // Current line is in use.
	else
		*pHeight = y; // Exclude current line.
	
	// We ended up here in our string.
	return pText;
}

void TEXT_DrawRect (uint16_t left, uint16_t top, uint16_t right, uint16_t bottom, uint16_t fill /* = 0xffff */)
{
	if (right  >= (left+2) &&
		bottom >= (top+2))
	{
		uint16_t colorMask = 0x8000;
		uint16_t x, y;

		uint16_t* offset = TEXT_RAM_BASE+(top*TEXT_SCREEN_WIDTH);
		offset[left] = colorMask | 1;
		for (x=left+1; x<right-1; x++)
			offset[x] = colorMask | 2;
		offset[right-1] = colorMask | 3;
		offset += TEXT_SCREEN_WIDTH;

		for (y=top+1; y<bottom-1; y++)
		{
			offset[left] = colorMask | 4;
			
			// Clear background (optional).
			if (fill != 0xffff)
				for (x=left+1; x<right-1; x++)
					offset[x] = fill;
					
			offset[right-1] = colorMask | 5;
			offset += TEXT_SCREEN_WIDTH;
		}

		for (x=left+1; x<right-1; x++)
		{
			offset[left] = colorMask | 6;
			for (x=left+1; x<right-1; x++)
				offset[x] = colorMask | 7;
			offset[right-1] = colorMask | 8;
		}
	}
}

void TEXT_DrawMessageWindow (const char* pText, bool bWordWrap)
{
	// Fixme: Should this store previous window state?
	uint16_t rectWidth = TEXT_SCREEN_VISIBLE_WIDTH-2, rectHeight = TEXT_SCREEN_HEIGHT-2;
	TEXT_CalcRect (&rectWidth, &rectHeight, pText, bWordWrap);
	
	uint16_t textX = TEXT_SCREEN_VISIBLE_XSTART + ((TEXT_SCREEN_VISIBLE_WIDTH - rectWidth) >> 1);
	uint16_t textY = ((TEXT_SCREEN_HEIGHT - rectHeight) >> 1);
	TEXT_DrawRect (textX-1, textY-1, textX+rectWidth+1, textY+rectHeight+1, ' ');
	
	TEXT_SetWindow (textX, textY, textX+rectWidth, textY+rectHeight, false);
	
	// Store old wrap state.
	bool oldWrap = s_windowState.AutoWordWrap;
	if (bWordWrap != oldWrap)
		s_windowState.AutoWordWrap = bWordWrap;

	TEXT_Write (pText);

	s_windowState.AutoWordWrap = oldWrap;
}

static const char hexTable[] = "0123456789ABCDEF";

// Write a hexadecimal number, up to 8 digits.
// Padding is prepended if the resulting string is shorter than minWidth.
void TEXT_WriteHex (uint32_t value, uint8_t minWidth, char padding /* = ' ' */)
{
	if (minWidth > 8)
		minWidth = 8;
	
	char hexString[9]; // Should be large enough for minWidth!
	const uint8_t maxChars = sizeof(hexString)-1;
	uint8_t writePos = maxChars;
	hexString[writePos] = '\0';
	
	while (value != 0)
	{
		uint8_t digit = value & 0xf;
		value >>= 4;
		hexString[--writePos] = hexTable[digit];
	}
	
	// Add at least one zero.
	if (writePos == maxChars)
		hexString[--writePos] = '0';
	
	while (maxChars-writePos < minWidth)
		hexString[--writePos] = padding;
	
	TEXT_Write (hexString + writePos);
}

void TEXT_WriteHex32 (uint32_t i)
{
	uint32_t shift = 32;
	do
	{
		shift -= 4;
		TEXT_WriteChar (hexTable[(i >> shift) & 0xf]);
	} while (shift);
}

void TEXT_WriteHex4 (uint8_t c)
{
	TEXT_WriteChar (hexTable[c & 0xf]);
}

void TEXT_WriteHex8 (uint8_t c)
{
	TEXT_WriteHex4 (c >> 4);
	TEXT_WriteHex4 (c);
}
