// Sample to demonstrate bi-directional communications over the PC (LPT or USB) interface.
//
// It'll print (and transmit) input actions received on the Out Run PCB into a console window on screen,
// yellow while still buffered and green when fully sent.
// It will also print any input received in a second console window on screen (eg. sent via orboot-usb)
//
// Note that this will require an active debug connection over the OutRun-USB interface, otherwise the buffer will
// never drain and the sample will be stuck after the first input action, staying yellow indefinitely.

#include "hwinit.h"
#include "palette.h"
#include "text.h"
#include "irq.h"
#include "io.h"
#include "comm.h"
#include "input.h"
#include <stdlib.h> // NULL
#include <string.h> // strlen

//------------------------------------------------------------------------------------------------------------------------------------------------

// Window IDs.
typedef enum
{
	kOutputWindow = 0,
	kInputWindow,
	kMAX_WINDOWS
} WindowID;

TEXT_WINDOW_STATE s_windowStates[kMAX_WINDOWS]; // Stored window states for both windows.
WindowID s_activeWindow; // Currently active window.

typedef struct _RECT
{
	uint16_t left, top, right, bottom;
} RECT;

// Set up and save initial window state.
// Paints border, title.
void SetupWindow (TEXT_WINDOW_STATE* pState, RECT rect, const char* title, uint8_t colorIdx)
{
	// Draw border.
	TEXT_DrawRect (rect.left, rect.top, rect.right, rect.bottom, 0xffff);

	// Draw caption.
	TEXT_SetWindow (rect.left, rect.top, rect.right, rect.top + 1, false);
	TEXT_GotoXY (1, 0);
	TEXT_SetColor (TEXT_Purple);
	TEXT_WriteChar (' ');
	TEXT_Write (title);
	TEXT_WriteChar (' ');

	RECT innerRect = rect; // Shrink.
	innerRect.top++; innerRect.bottom--;
	innerRect.left++; innerRect.right--;

	TEXT_SetWindow (innerRect.left, innerRect.top, innerRect.right, innerRect.bottom, true);
	TEXT_SetColor (colorIdx);
	TEXT_SetCursorEnable (true);

	TEXT_SaveWindowState (pState);
}

// Switch (with check) to a new window.
void SwitchWindow (WindowID window)
{
	if (window == s_activeWindow)
		return;
	TEXT_SaveWindowState (&s_windowStates[s_activeWindow]);
	TEXT_RestoreWindowState (&s_windowStates[window]);
	s_activeWindow = window;
}

//------------------------------------------------------------------------------------------------------------------------------------------------

COMM_CONTEXT s_commContext;

void Yield ()
{
	IRQ2_Wait();
}

void HandleRemoteInput ()
{
	if (!COMM_ReceiveAvailable (&s_commContext))
		return;

	SwitchWindow (kInputWindow);

	// Note: this could do longer buffers, but then we'd need to calculate the length in TEXT_Write.
	uint8_t b;
	while (COMM_Receive (&s_commContext, &b, 1))
	{
		if (b == 8)
		{
			// Backspace.
			uint8_t x, y;
			TEXT_GetXY (&x, &y);
			if (x > 0) 
			{
				x--; // Move back one position.
				TEXT_GotoXY (x, y);
				TEXT_WriteChar (' ');
				TEXT_GotoXY (x, y);
			}
			continue;
		}

		// Regular character forwarding (anything else).
		TEXT_WriteChar (b);
	}
}

void PrintAndRouteLocalInput (const char* buttonText)
{
	SwitchWindow (kOutputWindow);

	// If not at the left edge, add space.
	uint8_t x, y;
	TEXT_GetXY (&x, &y);
	if (x > 0)
		TEXT_WriteChar (' ');

	// Write a space between output, but since we don't necessarily (control) word-wrap on the host end, only skip space before the very first word.
	static bool isFirst = true;
	int addSpace = isFirst ? 0 : 1;
	isFirst = false;

	// Check if there's room for this string in the output buffer.
	int buttonTextLen = strlen (buttonText);
	uint8_t sendFree, sendBuffered;
	COMM_GetSendBufferState (&s_commContext, &sendBuffered, &sendFree);

	// If there's no room, draw it in red and wait.
	// Right now this never happens because we also wait for the buffer to completely drain, and the buffer is larger than any of the strings.            
	if (sendFree < (buttonTextLen + addSpace))
	{
		// No space in buffer.
		TEXT_SetColor (TEXT_Red);
		TEXT_Write (buttonText);
		TEXT_GetXY (&x, &y);
		x -= buttonTextLen;

		do
		{
			HandleRemoteInput();

			// Don't spam the IRQ lock.
			Yield();

			COMM_GetSendBufferState (&s_commContext, &sendBuffered, &sendFree);
		} while (sendFree < (buttonTextLen + addSpace));

		SwitchWindow (kOutputWindow);
		TEXT_GotoXY (x, y);
	}

	// Now there's room, so add it to the outgoing buffer.
	if (addSpace)
	{
		static const char space = ' ';
		COMM_Send (&s_commContext, &space, 1);
	}
	COMM_Send (&s_commContext, buttonText, buttonTextLen);

	// Write text in yellow.
	TEXT_SetColor (TEXT_Yellow);
	TEXT_Write (buttonText);
	TEXT_GetXY (&x, &y);
	x -= buttonTextLen;

	// Wait for buffer to drain.
	do
	{
		HandleRemoteInput();

		// Don't spam the IRQ lock.
		Yield();

		COMM_GetSendBufferState (&s_commContext, &sendBuffered, &sendFree);
	} while (sendBuffered);

	// Overwrite text in green.
	SwitchWindow (kOutputWindow);
	TEXT_GotoXY (x, y);
	TEXT_SetColor (TEXT_Green);
	TEXT_Write (buttonText);
}

InputState inputState; // Local input.

// Processes local input, and produces buffered comm output.
void HandleLocalInput ()
{
	INPUT_Update (&inputState);

	// This matches the InputButton enum in input.h
	static const char* buttonNames[MAX_BUTTONS] = {
		NULL, // 0 is reserved.
		"Test",
		"Service",
		"Start",
		"Gear",
		"Turbo",
		"Coin1",
		"Coin2",
		"Left",
		"Right",
		"Accel",
		"Brake",
	};

	// Skip 'Test' as it's used for data now.
	for (int i = Button_Service; i < MAX_BUTTONS; i++)
	{
		if (INPUT_IsButtonPressedNoRepeat (&inputState, i))
		{
			const char* const buttonText = buttonNames[i];
			PrintAndRouteLocalInput (buttonText);
		}
	}
}

void main ()
{
	// Default hardware init.
	HW_Init (HWINIT_Default | HWINIT_Clear_Road_Palette, PACKED_COLOR(0x321));

	// Set up windows on screen.
	TEXT_InitDefaultPalette();
	RECT rcTop = { TEXT_SCREEN_VISIBLE_XSTART + 2, 0, TEXT_SCREEN_WIDTH - 1, TEXT_SCREEN_HEIGHT >> 1 };
	RECT rcBottom = { rcTop.left, (TEXT_SCREEN_HEIGHT >> 1) + 0, TEXT_SCREEN_WIDTH - 1, TEXT_SCREEN_HEIGHT };
	SetupWindow (&s_windowStates[kOutputWindow], rcTop, "Output", TEXT_Yellow);
	SetupWindow (&s_windowStates[kInputWindow], rcBottom, "Input", TEXT_Cyan);

	s_activeWindow = kInputWindow;
	TEXT_RestoreWindowState (&s_windowStates[s_activeWindow]);

	// Initialize local input
	// Poll once to set up initial 'pressed' states so we don't get gear firing without actual change. 
	INPUT_Init (&inputState, true); // Analog also on IRQ2. 
	INPUT_Update (&inputState);

	// Setup comm (LPT) dispatcher.
	COMM_Init (&s_commContext, DIGITAL_OUT_MUTE_INTERNAL, true);
//	COMM_EnableReceive (&s_commContext, true);

	// Process any local input and send it to the host; write any remote input to the output window.
	for (;;)
	{
		// Handles local (user) input via analog/digital controls, and sends it out as text.
		HandleLocalInput ();
		HandleRemoteInput ();
	}
}
