#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include "palette.h"
#include "text.h"
#include "tile.h"
#include "sprite.h"
#include <io.h>
#include <irq.h>
#include "memtest.h"
#include "hwinit.h"
#include <stdlib.h>

// CRC-32 Functions
static void BuildCrcTable (unsigned long *crc32table)
{
    unsigned long c;
    int n, k;
    unsigned long poly;                 /* polynomial exclusive-or pattern */

    static const unsigned char p[] = {0,1,2,4,5,7,8,10,11,12,16,22,23,26};

	/* make exclusive-or pattern from polynomial (0xedb88320UL) */
    poly = 0xedb88320UL;
//    for (n = 0; n < sizeof(p)/sizeof(unsigned char); n++)
//        poly |= 1UL << (31 - p[n]);

	/* generate a crc for every 8-bit value */
    for (n = 0; n < 256; n++) {
        c = (unsigned long)n;
        for (k = 0; k < 8; k++)
            c = c & 1 ? poly ^ (c >> 1) : c >> 1;
        crc32table[n] = c;
    }
}

static unsigned long s_crcTable[256];

unsigned long CalcCRC32 (const void* pData, unsigned int nBytes, unsigned char nStride)
{
	unsigned long crc = 0xffffffff;
	unsigned char* pByteData = (unsigned char*)pData;
	while (nBytes)
	{
		crc = s_crcTable[((int)crc ^ (*pByteData)) & 0xff] ^ (crc >> 8);
		pByteData += nStride;
		--nBytes;
	}
	return crc ^ 0xffffffff;
}

// List of known ROMs.
const struct ROMINFO
{
	const char* name;
	unsigned short ic;
	unsigned long crc;
} romInfo[] = 
{
	// Main, Rev. B
	{ "EPR-10380b", 133, 0x1f6cadad },
	{ "EPR-10381b", 132, 0xbe8c412b },
	{ "EPR-10382b", 118, 0xc4c3fa1a },
	{ "EPR-10383b", 117, 0x10a2014a },

	// Sub-cpu.
	{ "EPR-10327a", 76,  0xe28a5baf },
	{ "EPR-10329a", 58,  0xda131c81 },
	{ "EPR-10328a", 75,  0xd5ec5e5d },
	{ "EPR-10330a", 57,  0xba9ec82a },
	
	{ "Empty.Mame",  0,  0xd7978eeb },
	{ "Empty.Real",  0,  0xdeab7e4e },
};

void RestorePalette (uint32_t baseAddr, bool succeeded)
{
	TEXT_InitDefaultPalette();
	
	// Clear road palette.
	const PackedColor color = PACKED_COLOR(0x000);
	for (uint16_t i=0x400; i<0x410; i++)
		PALETTE_SetColor (i, color);
	
	for (uint16_t i=0x420; i<0x440; i++)
		PALETTE_SetColor (i, color);

	// Background colors (scanline fill mode).
	for (uint16_t i=0; i<128; i++)
		PALETTE_SetColor (i+0x780, color);
}

// Ok, now I'm probably overengineering it a bit.
static bool s_bTileRamOk; // Updated by the tile ram callbacks.

// Called when the tile RAM chips are tested.
void TileCallback (uint32_t baseAddr, bool succeeded)
{
	if (!succeeded)
		s_bTileRamOk = false;
	
	if (baseAddr & 1)
	{
		TILE_Reset ();
		
		// Here we could copy away the contents of TEXT ram, if TILE ram is ok.
	}
}

void ResetResults ()
{
	s_bTileRamOk = true;
}

typedef void POSTCALLBACK(uint32_t baseAddr, bool succeeded);

const struct RAMINFO
{
	const char*    name;
	unsigned long  address;
	unsigned short interleave;
	unsigned long  size;
	POSTCALLBACK*  pRestoreCallback; // Called after the test.
	bool*          pConditionVar;    // Can point to a boolean that enables or disables the test.
} ramInfo[] = 
{
//	{ "Main0", 0x60000, 1, 2, 0x2000 },
//	{ "Main1", 0x60001, 1, 2, 0x2000 },
//	{ "Main2", 0x62000, 1, 2, 0x2000 },
//	{ "Main3", 0x62001, 1, 2, 0x2000 },


	// IC95 is een multiplexert/selector geval. SAB17..19 (1 t/m 19 op de CPU). LDS en UDS hebben we (byte select).
	// Ok dus CPU heeft gewoon geen A0. Nice. LDS=bit 0..7; UDS=bit 8..15.
	// 128k per rom bank (2 chips). Gebruikt bits 0..16 (op de M68k)
	// 
	//               LDS(DB 0..7)   UDS(DB 8..15)
	// 0x0200000     IC58           IC76
	// 0x0220000     IC57           IC75
	// 0x0240000     IC56           IC74
	// Als A17 en A18 aan (0x60000), dan gaan we naar RAM. A14 hoog -> !RAM1 hoog. !RAM1 -> !CE -> IC114 aan.
	//
	// Subram = 2x8k, dus extra selector op SAB14.
	// 0x0260000     IC54           IC72
	// 0x0220000     IC55(SA14=CE)  IC73(SA14=CE)	// Inverter (LS04 IC??) sets !CE to the inverse.

	// Big endian 16 bit data locations:
	// A 16-bit read to address #0 will load the high (first) byte at position 0 over lines D8..15, and the low byte at position 1 from lines D0..7.

	// SUB RAM.
	// 54 is the lower half due to the inverter (LS04) A14 to CE.
	// *** Note that sub ram needs to be untouched by the secondary CPU (including interrupts) during testing! ***
	{ "Sub IC72",  0x260000, 2, 0x4000 }, // 32k sub ram; 4 chips. ic 73+72(db8..15); 55+54(db0..7); 
	{ "Sub IC54",  0x260001, 2, 0x4000 },
	{ "Sub IC73",  0x264000, 2, 0x4000 },
	{ "Sub IC55",  0x264001, 2, 0x4000 },

	{ "Til IC64",  0x100000, 2, 0x10000 }, // 16 x 64*32 * 2 = 64k
	{ "Til IC62",  0x100001, 2, 0x10000, TileCallback },
//	{ "Text0   ",  0x110000, 2, 0x1000, NULL, &s_bTileRamOk }, // 4K text ram
//	{ "Text1   ",  0x110001, 2, 0x1000, NULL, &s_bTileRamOk },
	{ "Pal IC92",  0x120000, 2, 0x2000 }, // 8KB palette ram. Kun je hier uberhaupt wel halve words uit lezen? Of fixt die 68k dat intern?
	{ "Pal IC95",  0x120001, 2, 0x2000, RestorePalette },

	// Fixme: check IC #s.
	{ "Spr0    ",  0x130000, 2, 0x800 }, // Has a flip control, so this only tests half (could flip in a callback)
	{ "Spr1    ",  0x130001, 2, 0x800 },
};

static const char* itoa2 (int i, char* buf, size_t radix)
{
	bool isNegative = i < 0;
	uint32_t positive;
	if (isNegative)
		positive = ((i ^ 0xffffffff) + 1);
	else positive = i;
	
	char* write = buf + sizeof(buf);
	*(--write) = '\0';
	
	do
	{
		// Use 'div' to get both the quotient and the remainder.
		div_t result = div (positive, radix);
		*(--write) = '0' + result.rem;
		positive = result.quot;
	} while (positive);
	
	if (isNegative)
		*(--write) = '-';
	return write;
}

void ValidateRom (const void* address, unsigned short ic)
{
	unsigned int i;
	unsigned long crc;
	
	TEXT_SetColor (TEXT_White);

	char buf[10];
	TEXT_Write ("IC ");
	TEXT_Write (itoa2(ic, buf, 10));
	TEXT_Write (" ");

	crc = CalcCRC32 (address, 65536, 2);
	for (uint16_t i=0; i<sizeof(romInfo)/sizeof(struct ROMINFO); i++)
	{
		if (crc == romInfo[i].crc)
		{
			// Found a match.
			TEXT_SetColor ((romInfo[i].ic == ic) ? TEXT_Green : romInfo[i].ic ? TEXT_Red : TEXT_Purple);
			TEXT_Write (romInfo[i].name);
			TEXT_Write ("\n");
			return;
		}
	}

	// Not found.
	TEXT_SetColor (TEXT_Gray);
	TEXT_WriteHex32 (crc);
	TEXT_Write ("\n");
}
		
int main (void)
{
	HW_Init (HWINIT_Default, PACKED_COLOR(0x000));

	TEXT_InitDefaultPalette();
	TEXT_SetWindow (26, 1, 45, 28, false);
	TEXT_SetCursorEnable (true);

	BuildCrcTable (s_crcTable);
	
	for (;;)
	{
		// Clear variables.
		ResetResults ();
		
		TEXT_SetColor (TEXT_Cyan);
		TEXT_Write ("ROM TEST\n\n");

		ValidateRom ((void*)0x0, 133);
		ValidateRom ((void*)0x1, 118);
		ValidateRom ((void*)0x20000, 132);
		ValidateRom ((void*)0x20001, 117);
		ValidateRom ((void*)0x40000, 131);
		ValidateRom ((void*)0x40001, 116);

		TEXT_SetWindow (45, 3, 64, 28, false);
		ValidateRom ((void*)0x200000, 76);
		ValidateRom ((void*)0x200001, 58);
		ValidateRom ((void*)0x220000, 75);
		ValidateRom ((void*)0x220001, 57);
		ValidateRom ((void*)0x240000, 74);
		ValidateRom ((void*)0x240001, 56);

		TEXT_SetWindow (26, 1, 64, 28, true);

		// RAM TEST
		TEXT_GotoXY (0, 9);
		TEXT_SetColor (TEXT_Cyan);
		TEXT_Write ("RAM TEST\n\n");

		for (uint16_t i=0; i<sizeof(ramInfo)/sizeof(struct RAMINFO); i++)
		{
			// Check condition variable.
			const struct RAMINFO* pRamInfo = &ramInfo[i];
			if (pRamInfo->pConditionVar && !*pRamInfo->pConditionVar)
				continue; // Skip this test.
			
			bool bFailed = false;
			TEXT_SetColor (TEXT_White);
			TEXT_Write (ramInfo[i].name);
			TEXT_Write (" DB ");
			datum dbTest = memTestDataBus8 ((volatile datum*)ramInfo[i].address);
			if (dbTest)
			{
				TEXT_SetColor (TEXT_Red);
				TEXT_Write ("ERROR ");
				TEXT_SetColor (TEXT_Yellow);
				TEXT_WriteHex((int)dbTest, 2, '0');
				
				bFailed = true;
			}
			else 
			{
				TEXT_SetColor (TEXT_Green);
				TEXT_Write ("OK");
			}

			if (!bFailed)
			{
				TEXT_SetColor (TEXT_White);
				TEXT_Write (" AB ");
				datum* addrTest = memTestAddressBus ((volatile datum*)ramInfo[i].address, ramInfo[i].size, 2); // Fixme: grab from list.
				if (addrTest)
				{
					TEXT_SetColor (TEXT_Red);
					TEXT_Write ("ERROR ");
					TEXT_SetColor (TEXT_Yellow);
					TEXT_WriteHex((int)addrTest, 6, '0');

					bFailed = true;
				}
				else 
				{
					TEXT_SetColor (TEXT_Green);
					TEXT_Write ("OK");
				}
			}

			if (!bFailed)
			{
				TEXT_SetColor (TEXT_White);
				TEXT_Write (" DEV ");

				datum* devTest = memTestDevice ((volatile datum*)ramInfo[i].address, ramInfo[i].size, ramInfo[i].interleave);
				if (devTest)
				{
					TEXT_SetColor (TEXT_Red);
					TEXT_Write ("ERROR ");
					TEXT_SetColor (TEXT_Yellow);
					TEXT_WriteHex((int)devTest, 6, '0');
					bFailed = true;
				}
				else 
				{
					TEXT_SetColor (TEXT_Green);
					TEXT_Write ("OK");
				}
			}

			// Clean up what we just broke.
			if (ramInfo[i].pRestoreCallback)
				ramInfo[i].pRestoreCallback (ramInfo[i].address, !bFailed);
			
			TEXT_Write ("\n");
		}

		// 10 seconds to restart...
		for (int16_t countdown = 9; countdown >= 0; countdown--)
		{
			char buf[10];
			TEXT_SetColor (TEXT_Red);
			TEXT_GotoXY (0, 24);
			TEXT_Write ("Restarting in ");
			TEXT_Write (itoa2 (countdown, buf, 10));
			TEXT_Write ("...");
			
			// You should think about it, take a second.
			for (int16_t i=0; i<60; i++)
				IRQ4_Wait();
		}
		
		TEXT_ClearTextRAM ();
		TEXT_GotoXY (0, 0);
		
		// And loop.

	} // for (;;)
}
