#include "hwinit.h"
#include "text.h"
#include <sprite.h>
#include <irq.h>
#include <string.h> // memset
#include <palette.h>
#include <input.h>
#include <menu.h>
#include <io.h>
#include "palettes.h"
#include "spriteinfo.h"

// This copy will be filled by the user parameters, and then copied into the active sprite list.
SpriteData g_userSpriteData = { 0 };

// Default positions. This is the top left of where we're going to put the sprite.
static const uint16_t kSpriteTop  = 0x100 + 24;
static const uint16_t kSpriteLeft = 0x155 + 20 - 32;

const uint8_t legendX = 4, legendY = 22;

// Fills in the above structure with settings from one of the entries found by walking the sprite ROM banks.
void SetupSpriteDefaults (SpriteData* pSprite, uint16_t spriteIdx)
{
	const SpriteInfo* pSpriteInfo = &GameSprites[spriteIdx];
	pSprite->Bank            = pSpriteInfo->BankIdx;
	pSprite->BankOffset      = pSpriteInfo->BankOffset;
	pSprite->Pitch           = pSpriteInfo->Pitch;
	pSprite->Height          = pSpriteInfo->Height - 1;
	pSprite->Left            = kSpriteLeft;
	
	pSprite->DrawTopToBottom = 1;
	pSprite->FlipHorizontal  = 1;
	pSprite->DrawLeftToRight = 1;

	pSprite->HorizontalZoomFactor = 0x200;
	pSprite->VerticalZoomFactor   = 0x200;
	
	pSprite->PaletteIdx      = pSpriteInfo->PaletteIdx;
}

static void UpdateSpritePos (SpriteData* pSpriteData)
{
	if (!pSpriteData->DrawLeftToRight)
		pSpriteData->Left = kSpriteLeft + (pSpriteData->Pitch << 3);
	else
		pSpriteData->Left = kSpriteLeft;

	if (!pSpriteData->DrawTopToBottom)
		pSpriteData->Top = kSpriteTop + (pSpriteData->Height + 1);
	else
		pSpriteData->Top = kSpriteTop;
}

uint16_t g_spriteIdx = 0;

static void TEXT_DrawOnOff (bool bOn)
{
	TEXTGLYPH gl = bOn ? 9 : 11;
	TEXT_WriteRawChar (gl++);
	TEXT_WriteRawChar (gl++);
}

void PrintSpriteValues (const SpriteData* pSprite, uint16_t spriteIdx)
{
	const uint8_t xpos = 14;
	uint8_t ypos = 3;
	TEXT_SetColor (TEXT_Purple);
	
	TEXT_GotoXY (xpos-1, ypos);
	TEXT_WriteHex (spriteIdx, 3, ' ');

	// Palette
	ypos += 2;
	TEXT_GotoXY (xpos, ypos);
	TEXT_WriteHex (pSprite->PaletteIdx, 2, ' ');

	// Pitch
	ypos += 2;
	TEXT_GotoXY (xpos, ypos);
	TEXT_WriteHex ((uint8_t)pSprite->Pitch, 2, ' ');
	
	// Height
	ypos += 2;
	TEXT_GotoXY (xpos, ypos);
	TEXT_WriteHex (pSprite->Height, 2, ' ');

	TEXT_SetColor (TEXT_White);
	
	// Flip horizontal.
	ypos += 2;
	TEXT_GotoXY (xpos, ypos);
	TEXT_DrawOnOff (!pSprite->DrawLeftToRight);

	// Flip vertical.
	ypos += 2;
	TEXT_GotoXY (xpos, ypos);
	TEXT_DrawOnOff (!pSprite->DrawTopToBottom);

	TEXT_SetColor (TEXT_Purple);
	// Zoom X
	ypos += 2;
	TEXT_GotoXY (xpos-1, ypos);
	TEXT_WriteHex (pSprite->HorizontalZoomFactor, 3, ' ');
	// Zoom Y
	ypos += 2;
	TEXT_GotoXY (xpos-1, ypos);
	TEXT_WriteHex (pSprite->VerticalZoomFactor, 3, ' ');

	// Shadows.
	TEXT_SetColor (TEXT_White);
	ypos += 2;
	TEXT_GotoXY (xpos, ypos);
	TEXT_DrawOnOff (pSprite->EnableShadows);
	
	TEXT_SetColor (TEXT_Purple);
	
	// Print next sprite address (temp).
	//xpos = 10;
	int8_t valueX = legendX + 5;
	ypos = legendY;
	TEXT_GotoXY (valueX, ypos++);
	TEXT_SetColor (TEXT_Yellow);
	TEXT_WriteHex ((uint16_t)pSprite->Bank, 1, '0');
	TEXT_GotoXY (valueX, ypos++);
	TEXT_WriteHex ((uint16_t)pSprite->BankOffset, 4, '0');
	TEXT_GotoXY (valueX, ypos++);
	TEXT_WriteHex ((uint16_t)(pSprite->BankOffset + pSprite->Pitch * (pSprite->Height + 1)), 4, '0');
}

void ChangeSprite (InputButton button, void* param)
{
	bool spriteChanged = false;
	if (button == Button_Brake && g_spriteIdx > 0)
	{
		g_spriteIdx--;
		spriteChanged = true;
	} 
	else if (button == Button_Accelerate && (g_spriteIdx+1) < NumGameSprites)
	{
		g_spriteIdx++;
		spriteChanged = true;
	}
	
	if (spriteChanged)
	{
		SetupSpriteDefaults (&g_userSpriteData, g_spriteIdx);
		PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
	}
}

void ChangePitch (InputButton button, void* param) 
{
	if (g_userSpriteData.Pitch > -64 && button == Button_Brake)
		g_userSpriteData.Pitch--;
	else if (g_userSpriteData.Pitch < 63 && button == Button_Accelerate)
		g_userSpriteData.Pitch++;
	
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

void ChangeHeight (InputButton button, void* param) 
{
	if (g_userSpriteData.Height > 0 && button == Button_Brake)
		g_userSpriteData.Height--;
	else if (g_userSpriteData.Height < 0xff && button == Button_Accelerate)
		g_userSpriteData.Height++;
	
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

// Since this is a partial (10 bit) uint16, we can't easily the the address and need to duplicate.
void ChangeZoomHorizontal (InputButton button, void* param)
{
	if (g_userSpriteData.HorizontalZoomFactor > 0 && button == Button_Brake)
		g_userSpriteData.HorizontalZoomFactor--;
	else if (g_userSpriteData.HorizontalZoomFactor < 0x3ff && button == Button_Accelerate)
		g_userSpriteData.HorizontalZoomFactor++;
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

void ChangeZoomVertical (InputButton button, void* param)
{
	if (g_userSpriteData.VerticalZoomFactor > 0 && button == Button_Brake)
		g_userSpriteData.VerticalZoomFactor--;
	else if (g_userSpriteData.VerticalZoomFactor < 0x3ff && button == Button_Accelerate)
		g_userSpriteData.VerticalZoomFactor++;
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

void ChangePalette (InputButton button, void* param) 
{
	if (g_userSpriteData.PaletteIdx > 0 && button == Button_Brake)
		g_userSpriteData.PaletteIdx--;
	else if (g_userSpriteData.PaletteIdx < 127 && button == Button_Accelerate)
		g_userSpriteData.PaletteIdx++;
	
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

static void FlipHorizontal (InputButton button, void* param) 
{
	g_userSpriteData.DrawLeftToRight ^= 1; // This one just really draws left to right. No address change required.
	UpdateSpritePos (&g_userSpriteData); // Update position.
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

static void FlipVertical (InputButton button, void* param) 
{
	// This just reverses the vertical paint direction.
	g_userSpriteData.DrawTopToBottom ^= 1;

	UpdateSpritePos (&g_userSpriteData); // Update position, this handles DrawTopToBottom.
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

static void ToggleShadows (InputButton button, void* param) 
{
	g_userSpriteData.EnableShadows ^= 1;
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
}

static const MenuItem s_menuItems[] = 
{
	{ "Sprite",  ChangeSprite, 0 },
	{ "Palette", ChangePalette, 0 },
	{ "Pitch",   ChangePitch, 0 },
	{ "Height",  ChangeHeight, 0 },
	{ "Flip H",  FlipHorizontal, 0 }, // DrawLeftToRight
	{ "Flip V",  FlipVertical, 0 },   // DrawTopToBottom
	{ "Zoom H",  ChangeZoomHorizontal, 0 },
	{ "Zoom V",  ChangeZoomVertical, 0 },
	{ "Shadows", ToggleShadows, 0 },
};

// Any text that doesn't change when values update; title, legend, usage.
void PrintStaticText ()
{
	// Print header.
	TEXT_GotoXY (14,1);
	TEXT_SetColor (TEXT_Yellow);
	TEXT_Write ("sprite sample");

	// Instructions. Left/right to select, Accelerate to play.
	TEXT_GotoXY (6,26);
	TEXT_SetColor (TEXT_White);
	TEXT_WriteRawChar (TEXTGLYPH_ARROWLEFT);
	TEXT_Write (" \006or\010 ");
	TEXT_WriteRawChar (TEXTGLYPH_ARROWRIGHT);
	TEXT_Write ("\006 select  \010");
	TEXT_WriteRawChar (TEXTGLYPH_ARROWUP);
	TEXT_Write (" \006or\010 ");
	TEXT_WriteRawChar (TEXTGLYPH_ARROWDOWN);
	TEXT_Write ("\006 modify");

	// Draw 'Bank', 'Addr' etc.
	TEXT_SetColor (TEXT_White);
	TEXT_GotoXY(legendX, legendY);
	TEXT_Write ("Bank ");
	TEXT_GotoXY(legendX, legendY+1);
	TEXT_Write ("Addr ");
	TEXT_GotoXY(legendX, legendY+2);
	TEXT_Write ("Next ");
}

void main ()
{
	HW_Init (HWINIT_Default, PACKED_COLOR(0x321)); // Very dark blue.
	TEXT_InitDefaultPalette ();

	PrintStaticText();

	// Initialize input helper.
	InputState inputState;
	INPUT_Init (&inputState, true);

	// Initialize menu, this draws the initial menu state.
	MenuInitParams initParams;
	initParams.iLeft = 2; initParams.iTop = 3;
	MenuState menuState;
	MENU_Init (&menuState, s_menuItems, sizeof(s_menuItems)/sizeof(MenuItem), &initParams, MIF_POSITION);

	SetupSpritePalettes ();
	
	// Globals.	
	g_userSpriteData.Top = 0x100 + 24;
	g_userSpriteData.Left = 0x155 + 20 - 32;
	g_userSpriteData.HorizontalZoomFactor = 0x200;
	g_userSpriteData.VerticalZoomFactor = 0x200;

	SetupSpriteDefaults (&g_userSpriteData, g_spriteIdx);
	PrintSpriteValues (&g_userSpriteData, g_spriteIdx);
	
	for (;;)
	{
		static uint8_t x = 0;
		x++;

		// Poll input.
		INPUT_Update (&inputState);
		MENU_Update (&menuState, &inputState);

		SpriteData userSprite = g_userSpriteData; // copy.
		SpriteData* pSprite = &userSprite; // Fill this thing.
		
		pSprite->Priority = 3;
		
		// Now write to buffer.
		SpriteList[0] = *pSprite;
		*((uint16_t*)&SpriteList[1]) = 0xffff; // TERMINATE
		
		IRQ4_Wait ();
		SPRITE_SwapBuffers ();
	}
}
