#include "hwinit.h"
#include "text.h"
#include "tile.h"
#include "palette.h"
#include <irq.h>
#include <stdint.h>
#include <stdlib.h>
#include "tileunpack.h"
#include "input.h"
#include "paletteutil.h"
#include "road.h"

extern const TileGraphics CloudGraphics;
extern const TileGraphics ShoreGraphics;

void SetupGradient ()
{
	const uint8_t colorStart[3] = { 113,48,243 };
	const uint8_t colorEnd[3]   = { 243,178,186 }; // From le website.
	
	uint16_t red   = (uint16_t)colorStart[0] << 8;
	uint16_t green = (uint16_t)colorStart[1] << 8;
	uint16_t blue  = (uint16_t)colorStart[2] << 8;
	
	int16_t d_red   = (colorEnd[0] - colorStart[0]) << 1;
	int16_t d_green = (colorEnd[1] - colorStart[1]) << 1;
	int16_t d_blue  = (colorEnd[2] - colorStart[2]) << 1;
	
	int16_t err_red = 0, err_green = 0, err_blue = 0;
	
	for (uint8_t i=0; i<128; i++)
	{
		int16_t newRed = red + err_red;
		int16_t newGreen = green + err_green;
		int16_t newBlue = blue + err_blue;
		
		err_red = newRed & 0x7ff;
		err_green = newGreen & 0x7ff;
		err_blue = (newBlue & 0x7ff);
		
		PALETTE_SetColorRGB (0x780+i, newRed >> 11, newGreen >> 11, newBlue >> 11);
		
		red   += d_red;
		green += d_green;
		blue  += d_blue;
	}

	// Set up the road data to match.
	const uint8_t grad_start = 87;
	uint8_t y;
	for (y=0; y<grad_start; y++)
		ROAD_PATTERN_0[y] = ROAD_SOLID_COLOR(0);
	for (y=grad_start; y<grad_start+127; y++)
		ROAD_PATTERN_0[y] = ROAD_SOLID_COLOR(y-grad_start);
	for (y=grad_start+127; y<224; y++)
		ROAD_PATTERN_0[y] = ROAD_SOLID_COLOR(127);
	
	// Swap buffers.
	ROAD_SwapBuffers ();
}

// Basic 5-digit itoa.
static void PrintScrollValue (uint16_t val)
{
	char digits[6] = { "     " };
	uint8_t pos = sizeof(digits)-2;
	while (val)
	{
		div_t result = div (val, 10);
		uint8_t digit = result.rem;
		val = result.quot;
		if (digit || val)
			digits[pos--] = '0'+digit;
	}
	TEXT_Write (digits);
}

// Used tile pages.
static const uint8_t kForegroundStartPage = 4;
static const uint8_t kBackgroundStartPage = 8;
static const uint8_t kEmptyPage           = 0; // Cleared by HW_Init.

struct ScrollInfo
{
	// Constants.
	const uint8_t const * const         pTilePages;       // List of pages to scroll through.
	const uint8_t                       nPageCount;       // Number of pages.
	volatile struct TILE_XSCROLL* const pScrollRegisterX; // Destination register for X position.
	volatile struct TILE_PAGES* const   pScrollPages;	  // Visible page indices.
	
	// State variables.
	uint16_t 					  iScrollPosX;		// Current scroll position (as written to x scroll register).
	uint8_t					 	  iFocusPage;		// Page the center of the viewport currently peeks into.
};

// Order in which the pages should tile, left-to-right.
static const uint8_t kBackgroundPages[] = { 8, 9, 10 }; // Clouds.
static const uint8_t kForegroundPages[] = { 4, 5, 6, 7 }; // Water, trees, etc.

// Fill in constants.
static struct ScrollInfo s_scrollTargets[] = 
{
	{
		kBackgroundPages,
		sizeof(kBackgroundPages),
		&TileRegisters.BackgroundScrollX,
		&TileRegisters.BackgroundMapPages,
		0, // X position; past the border.
		sizeof(kBackgroundPages)-1,
	},
	{
		kForegroundPages,
		sizeof(kForegroundPages),
		&TileRegisters.ForegroundScrollX,
		&TileRegisters.ForegroundMapPages,
		0,
		sizeof(kForegroundPages)-1,
	},
};

// Set up initial values and register contents.
void InitScrollTarget (struct ScrollInfo* pScrollInfo)
{
	pScrollInfo->iScrollPosX = 0;
	pScrollInfo->iFocusPage  = (pScrollInfo->nPageCount-1);
	
	uint16_t visiblePages = 
		(kEmptyPage << 12) | (kEmptyPage << 8) | 
		((pScrollInfo->pTilePages[(pScrollInfo->iFocusPage + 1) % pScrollInfo->nPageCount]) << 4) | 
		  pScrollInfo->pTilePages[ pScrollInfo->iFocusPage];
		  
	pScrollInfo->pScrollRegisterX->X = pScrollInfo->iScrollPosX;
	pScrollInfo->pScrollPages->Pages = visiblePages;
}

// Update horizontal scroll position, and swaps in appropriate pages when needed.
void UpdateScrollTarget (struct ScrollInfo* pScrollInfo, int16_t scrollSpeedX)
{
	// Update scroll position. Value in the struct is 12:4 fixed point.
	uint16_t prevScrollPosX = (pScrollInfo->iScrollPosX >> 6) & 1023;
	pScrollInfo->iScrollPosX += scrollSpeedX;
	uint16_t scrollPosX = (pScrollInfo->iScrollPosX >> 6) & 1023; // Shift down and keep 10 bits.
	
	// Whenever we near the edge of the page, we set up the adjoining page.
	if (((scrollPosX ^ prevScrollPosX) & 512) == 0) // Only update when still on the same page.
	{
		if (pScrollInfo->iFocusPage != 0xff) // Center moving to edges.
		{
			uint8_t nextFocusPage = 0xff;
			if ((scrollPosX & 511) >= 144 && (prevScrollPosX & 511) < 144)
			{
				// Nearing the left border. Flip in the new page.
				nextFocusPage = pScrollInfo->iFocusPage;
				if (nextFocusPage-- == 0)
					nextFocusPage = (pScrollInfo->nPageCount-1);
			}
			else if ((scrollPosX & 511) < 48 && (prevScrollPosX & 511) >= 48)
			{
				// Nearing the right border.
				nextFocusPage = pScrollInfo->iFocusPage;
				if (++nextFocusPage == pScrollInfo->nPageCount)
					nextFocusPage = 0;
			}
			
			if (nextFocusPage != 0xff)
			{
				// Set up 'newFocusPage' on the right side, which is the page we're not in.
				if (scrollPosX & 512)
				{
					// Focus on left, update right page (lowest nibble)
					pScrollInfo->pScrollPages->Page_0 = pScrollInfo->pTilePages[nextFocusPage];
				}
				else
				{
					// Focus on right, update left page (second nibble)
					pScrollInfo->pScrollPages->Page_1 = pScrollInfo->pTilePages[nextFocusPage];
				}
			}
		}

		// If the center of the viewport crosses a page border, update the focus page.
		// Threshold = 512 - (320/2) = 352.
		if ((scrollPosX & 511) < 352 && (prevScrollPosX & 511) >= 352)
		{
			// Moving right.
			if (++pScrollInfo->iFocusPage == pScrollInfo->nPageCount)
				pScrollInfo->iFocusPage = 0;
		}
		else if ((prevScrollPosX & 511) < 352 && (scrollPosX & 511) >= 352)
		{
			// Moving left.
			if (pScrollInfo->iFocusPage-- == 0)
				pScrollInfo->iFocusPage = pScrollInfo->nPageCount-1;
		}
	}

	// Update scroll register.
	pScrollInfo->pScrollRegisterX->X = scrollPosX;
}

void PrintScrollInfo (const struct ScrollInfo* pScrollInfo)
{
	// Draws the pages currently visible, in different colors:
	// Gray:  Completely off screen.
	// Blue:  Partially visible (clipped).
	// Cyan:  Partially visible, and the center of the viewport is on this page.
	// White: Viewport is completely contained within this page. 
	TEXT_Write ("\002 PAGES ");

	// Add some fancy colors.
	uint16_t x = pScrollInfo->pScrollRegisterX->X;
	if (x > 512 && x < 192+512)
		TEXT_SetColor (TEXT_White); // Viewport completely inside page.
	else if (x < 192)
		TEXT_SetColor (TEXT_Gray);  // Completely outside page.
	else if (x >= 352 && x <= (352+512))
		TEXT_SetColor (TEXT_Cyan); // At least the center is inside.
	else 
		TEXT_SetColor (TEXT_Blue);
	TEXT_WriteHex4 (pScrollInfo->pScrollPages->Page_1);
	
	if (x < 192)
		TEXT_SetColor (TEXT_White); // Viewport completely inside page.
	else if (x >= 512 && x < 512+192)
		TEXT_SetColor (TEXT_Gray);  // Completely outside page.
	else if (x <= 352 || x >= (352+512))
		TEXT_SetColor (TEXT_Cyan); // At least the center is inside.
	else TEXT_SetColor (TEXT_Blue);
	TEXT_WriteHex4 (pScrollInfo->pScrollPages->Page_0);
	
	TEXT_Write ("\002 FOCUS \006");
	TEXT_WriteHex4 (pScrollInfo->pTilePages[pScrollInfo->iFocusPage]);
	
	TEXT_Write ("\002 X\006");
	PrintScrollValue (pScrollInfo->pScrollRegisterX->X);
}

// Draws the title and stats.
void DrawInfo (bool enableStats)
{
	static bool isFirstTime = true;
	static bool prevEnableStats = false;

	// 'Tile map sample'
	if (isFirstTime)
	{
		// Statically positioned on top, so only draw once.
		TEXT_GotoXY (13,1);
		TEXT_SetColor (TEXT_Yellow);
		TEXT_Write ("tile map sample");
	}

	// Stats, if enabled.
	if (enableStats)
	{
		for (uint8_t i=0; i<2; i++)
		{
			TEXT_GotoXY (3,3+i);
			static const char* pPrefixes[] = { "\010Background\006", "\010Foreground\006" }; // Their most accomplished album.
			TEXT_Write (pPrefixes[i]);
			PrintScrollInfo (&s_scrollTargets[i]);
		}
	}

	// Clear some lines, if needed.	
	if (enableStats != prevEnableStats)
	{
		TEXT_GotoXY (3, 3); 
		uint8_t nLines = prevEnableStats ? 4 : 1;
		for (uint8_t i=0; i<nLines; i++)
		{
			TEXT_ClearEOL(); TEXT_Write ("\n");
		}	
	}
	
	// Instructions. Left/right to select, Accelerate to play.
	if (enableStats != prevEnableStats || isFirstTime)
	{
		TEXT_GotoXY (7, enableStats ? 6 : 3);
		TEXT_SetColor (TEXT_White);
		TEXT_WriteRawChar (TEXTGLYPH_ARROWLEFT);
		TEXT_Write (" \006or\010 ");
		TEXT_WriteRawChar (TEXTGLYPH_ARROWRIGHT);
		TEXT_Write ("\006 to change direction  \010");
	}
	
	isFirstTime = false;
	prevEnableStats = enableStats;
}

// Speed constraints (8:8 fixed point)
static const int16_t kScrollSpeedMin = -1024;
static const int16_t kScrollSpeedMax =  1024;

void main ()
{
	HW_Init (HWINIT_Default, PACKED_COLOR(0xf63));
	TEXT_InitDefaultPalette ();
	
	InputState inputState;
	INPUT_Init (&inputState, true);
	
	// Unpack maps.
	UnpackTileGraphics (kBackgroundStartPage, &CloudGraphics);
	UnpackTileGraphics (kForegroundStartPage, &ShoreGraphics);
	
	// Set the color gradient.
	SetupGradient ();
	
	// Set up fixed Y positions.
	TileRegisters.BackgroundScrollY.Y =
	TileRegisters.ForegroundScrollY.Y = 4*8;

	// Initial register setup.	
	for (uint8_t i=0; i<2; i++)
		InitScrollTarget (&s_scrollTargets[i]);

	// Initial scroll speed.	
	int16_t scrollSpeedX = 2 << 8; // 5:8 fixed point.
		
	for (;;)
	{
		// Independent scroll layer updates.
		UpdateScrollTarget (&s_scrollTargets[0], scrollSpeedX >> 2); // Background (clouds)
		UpdateScrollTarget (&s_scrollTargets[1], scrollSpeedX >> 1); // Foreground (beach/trees/water)
		
		// Update scroll value.
		INPUT_Update (&inputState);
		int8_t inputAxis = (int8_t)(128-INPUT_GetAxisValue (&inputState, Axis_Steering));
		scrollSpeedX += (inputAxis >> 1);
		if (scrollSpeedX < kScrollSpeedMin)
			scrollSpeedX = kScrollSpeedMin;
		else if (scrollSpeedX > kScrollSpeedMax)
			scrollSpeedX = kScrollSpeedMax;

		// Display stats.
		bool enableStats = INPUT_IsButtonHeld (&inputState, Button_Gear);
		DrawInfo (enableStats);		
								
		IRQ4_Wait ();
	}
}
