#ifndef __PALETTE_H__
#define __PALETTE_H__

/*
	The palette consists of 8K ram = 4K 16-bit color entries.
	The bottom 12 bits constitute a 4:4:4 BGR color; bits 12-14 add an extra bit of precision.
	Bit 15 is used in highlight mode, where the background is brightened if the bit is set,
	or darkened when the bit is not set.
	Accessing palette ram during the active display period will cause palette noise.

	The first 64 colors are used as 8 sets of 8 for the text layer.
	The road layer uses the following color ranges: 0x400-0x40f, 0x420-0x43f and 0x780-0x7ff.
	Sprites use 128 16-color (4bpp) palettes, starting at 0x800.
*/

#include <stdint.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" 
{
#endif // __cplusplus

// Color layout.
typedef union
{
	uint16_t Packed; // Adding this first so we can use initializer lists, i.e. PackedColor cols[] = { 0x222, 0x123 };
	struct
	{
		uint16_t Shade : 1;
		
		// Extra precision bits.
		uint16_t BlueExtra  : 1; 
		uint16_t GreenExtra : 1;
		uint16_t RedExtra   : 1;

		// Base 4:4:4 color.
		uint16_t Blue  : 4;
		uint16_t Green : 4;
		uint16_t Red   : 4;
	};
} PackedColor;

// Helper to convert literal colors (eg. 0x123) to the PackedColor union type.
#define PACKED_COLOR(x) ((PackedColor)(uint16_t)x)

enum
{
	PALETTE_ROAD_START_IDX       = 0x400,
	PALETTE_ROAD_END_IDX         = 0x410,
	PALETTE_ROAD_BG_START_IDX    = 0x420,
	PALETTE_ROAD_BG_END_IDX      = 0x440,
	PALETTE_ROAD_SOLID_START_IDX = 0x780, // 128 colors for solid scanline fill mode.
	PALETTE_SPRITE_START_IDX     = 0x800, // 128 * 16 colors.
};

//  Use these instead.
void PALETTE_SetColor (uint16_t idx, const PackedColor packed);
PackedColor PALETTE_GetColor (uint16_t idx);

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif // __PALETTE_H__
