#ifndef __COMM_H__
#define __COMM_H__

#include "platform.h"

class CommBase
{
public:
	virtual bool Init() = 0;
	virtual void Reset() = 0;
	virtual ~CommBase () {} 

	// Sets the response timeout. Set to zero for no timeout.
	virtual void SetRXTimeOutMS (uint32 milliseconds) = 0;
	virtual void SetTXTimeOutMS (uint32 milliseconds) = 0;

	// Sends a single byte in 'compatibility mode'.
	// Returns false if the send times out.
	virtual bool SendByte (uint8 byte) = 0;

	// Reads a single byte in 'nibble' mode.
	// Returns -1 if the receive times out.
	virtual int16 RecvByte () = 0;
};

// Parallel protocol based on IEEE1284.
// Difference is that we actually wait for the transitions instead of relying on
// adequate 500ns response time. We don't have hooked up interrupts and 10MHz is too slow.
class CommLPT : public CommBase
{
public:
	// Optional parameters.
	struct Options
	{
		uint16 basePort;
		uint64 txDelayNS;
		uint32 debugDelayMS;

		// Defaults.
		Options() :
			basePort(0x378),
			txDelayNS(50000),
			debugDelayMS(0)
		{
		}
	};

	// Initializes timers. Run from the same thread as all the other calls.
	//void Init () override;
	CommLPT (const Options& options);

	// Set the transition delay. Default is 50ms to be safe. 35ms is about the minimum transition time
	// for the TLP521 optocouplers used on the Outrun board.
	void SetTXDelayNS (uint64 nanoseconds);

	// Sets a debug delay for all output transitions. Zero (default) means no delay.
	void SetDebugDelayMS (uint32 milliseconds);

	// Detect whether inpout32 driver is installed.
	bool Init () override;

	// Resets the control register:
	// - nSTROBE line is low (we inverted this ourselves!) (not strobing).
	// - nAUTOFEED line is high (HostBusy for nibble mode).
	// - Data bits are cleared.
	void Reset () override;

	// Sets the response timeout. Set to zero for no timeout.
	void SetRXTimeOutMS (uint32 milliseconds) override;
	void SetTXTimeOutMS (uint32 milliseconds) override;

	// Sets an inversion mask for the control register. Default is zero (normal behavior).
	void SetControlInversionMask (uint8 mask);

	// Sends a single byte in 'compatibility mode'.
	// Returns false if the send times out.
	bool SendByte (uint8 byte) override;

	// Reads a single byte in 'nibble' mode.
	// Returns -1 if the receive times out.
	int16 RecvByte () override;

}; // class CommLPT

#endif // __COMM_H__
