#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#if defined(_MSC_VER) || defined(__MINGW32__)
#include <malloc.h>
#else
#include <alloca.h>
#endif

// SplitBin - Splits binary images for use w/ split ROMs (eg. for 16 bit platorms using pairs of 8 bit EPROMs).
// 8/2019: Partial overhaul to make this cross platform and C(11) compatible (for potential linux/osx targets).

int main (int argc, char **argv)
{
	if (argc < 5)
	{
		printf ("SplitBin - Splits a binary image in to separate (interleaved) (EP)ROM images.\n");
		printf ("Usage:   splitbin <inputfile> <rom size> <interleave> <name1> [<name2> ...]\n");
		printf ("Example: splitbin maincpu.bin 65536 2 epr-10380b.133 epr-10382b.118\n");
		printf ("Note: The file names should be supplied in interleaved order.\n");
		return 0;
	}

	const char* inputFileName = NULL;
	unsigned int romSize = 0, interleave = 0;
	
	// Output files.
	const char** outputRomNames = (const char**)alloca(sizeof(const char*)*argc);
	size_t outputRomNameCount = 0; 

	// Process the command line.
	int argIdx = 1;
	while (argIdx < argc)
	{
		const char* arg = argv[argIdx];
		if (arg[0] == '-')
		{
			// Handle options here.
			printf ("Invalid option: '%s'.\n", arg);
			return 1;
		}
		else if (!inputFileName)
		{
			inputFileName = arg;
		}
		else if (!romSize)
		{
			if (sscanf (arg, "%u", &romSize) != 1 || romSize == 0)
			{
				printf ("Invalid ROM size: '%s'.\n", arg);
				return 1;
			}
		}
		else if (!interleave)
		{
			if (sscanf (arg, "%u", &interleave) != 1 || interleave == 0)
			{
				printf ("Invalid interleave value: '%s'.\n", arg);
				return 1;
			}
		}
		else // Output file names.
		{
			outputRomNames[outputRomNameCount++] = arg;
		}
		argIdx++;
	}

	if (outputRomNameCount == 0)
	{
		printf ("No output file names specified.\n");
		return 1;
	}

	if ((outputRomNameCount % interleave) != 0)
	{
		printf ("Mismatching number of output file names; expected a multiple of the interleave value.\n");
		return 1;
	}

	FILE* pInFile = fopen (inputFileName, "rb");
	if (!pInFile)
	{
		printf ("Input file '%s' not found.\n", inputFileName);
		return 1;
	}

	// Get file size; only used to verify the number of output files supplied is large enough.
	long maxFileSize = romSize * outputRomNameCount;
	long fileSize = -1L;
	if (fseek (pInFile, 0, SEEK_END) == 0)
	{
		fileSize = ftell (pInFile);
		fseek (pInFile, 0, SEEK_SET);
	}

	if (fileSize == -1L)
	{
		printf ("Can't get file size for input file '%s'.\n", inputFileName);
		return 1;
	}

	if (fileSize == 0)
	{
		printf ("Input file '%s' is empty, no ROMs written.\n", inputFileName);
		return 1;
	}

	if (fileSize > maxFileSize)
	{
		printf ("Not enough output file names to accomodate input file size of %u.\n", (unsigned int)fileSize);
		return 1;
	}

	printf ("Opened '%s', %u byte%s\n", inputFileName, (unsigned int)fileSize, fileSize==1?"":"s");

	char* sequentialData = (char*)malloc(romSize * interleave); // consecutive pre-split chunk.
	char* outputBuffer = (char*)malloc(romSize); // single output ROM used during write.
	const char** nextRomName = outputRomNames;

	for (;;)
	{
		size_t dataSize = fread (sequentialData, 1, (romSize*interleave), pInFile);

		// Todo: check ferror/feof.
		if (dataSize == 0)
			break; // No more data, we're done.

		// Write out a set.
		for (size_t j=0; j<interleave; j++)
		{
			size_t romUsage = 0;
			for (size_t i=j; i<dataSize; i+=interleave)
				outputBuffer[romUsage++] = sequentialData[i];
			memset (outputBuffer+romUsage, 0, romSize-romUsage); // Zero out the rest.

			const char* romName = *(nextRomName++);
			FILE* pOutFile = fopen (romName, "wb");
			if (!pOutFile)
			{
				printf ("Error creating output file '%s' (%u).", romName, errno);
				fclose (pInFile);
				return 1;
			}

			if (fwrite (outputBuffer, 1, romSize, pOutFile) != romSize)
			{
				printf ("Error writing to output file '%s' (%u).", romName, errno);
				fclose (pOutFile);
				fclose (pInFile);
				return 1;
			}

			fclose (pOutFile);
			printf ("Written '%s'.\n", romName);
		}

		if (dataSize < romSize*interleave)
			break; // upcoming EOF, done.
	}

	free (sequentialData);
	free (outputBuffer);
	fclose (pInFile);
	return 0;
}
